# Copyright (C) Stichting Deltares. All rights reserved.
#
# This file is part of the Probabilistic Library.
#
# The Probabilistic Library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# All names, logos, and references to "Deltares" are registered trademarks of
# Stichting Deltares and remain full property of Stichting Deltares at all times.
# All rights reserved.
#


import sys
from math import isnan
from enum import Enum
import matplotlib.pyplot as plt
import numpy as np

from .utils import FrozenObject, FrozenList, PrintUtils
from .statistic import Stochast, ProbabilityValue
from .reliability import StochastSettings, GradientType
from .logging import Evaluation, Message, ValidationReport
from . import interface

if not interface.IsLibraryLoaded():
	interface.LoadDefaultLibrary()

class SensitivityMethod(Enum):
	"""Enumeration which defines the algorithm to perform a sensitivity analysis"""
	single_variation = 'single_variation'
	sobol = 'sobol'
	def __str__(self):
		return str(self.value)

class SensitivitySettings(FrozenObject):
	"""Settings of a sensitivity algorithm

    Settings of all sensitivity algorithms are combined in this class. Often settings only apply to
    a selected number of algorithms. Settings per variable are listed in `stochast_settings`.

    The settings are divided into the following categories:

    Algorithm settings: Settings which are used to control the algorithm, such as variations of the input values.

    Runtime settings: Settings to control how model executions take place and which additional output to produce.
    These settings do not influence the calculated sensitivity."""

	def __init__(self):
		self._id = interface.Create('sensitivity_settings')
		self._stochast_settings = FrozenList()
		self._quantiles = None
		self._synchronizing = False
		super()._freeze()
		interface.SetBoolValue(self._id, 'use_openmp_in_reliability', False)

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['max_parallel_processes',
		        'save_realizations',
		        'save_convergence',
		        'save_messages',
		        'reuse_calculations',
		        'sensitivity_method',
		        'low_value',
		        'high_value',
		        'iterations',
		        'is_valid',
		        'validate']

	@property
	def max_parallel_processes(self) -> int:
		"""The number of parallel executions of model evaluations"""
		return interface.GetIntValue(self._id, 'max_parallel_processes')

	@max_parallel_processes.setter
	def max_parallel_processes(self, value : int):
		interface.SetIntValue(self._id, 'max_parallel_processes', value)

	@property
	def save_realizations(self) -> bool:
		"""Indicates whether samples should be saved
        If saved, the samples will be part of the design point"""
		return interface.GetBoolValue(self._id, 'save_realizations')

	@save_realizations.setter
	def save_realizations(self, value : bool):
		interface.SetBoolValue(self._id, 'save_realizations', value)

	@property
	def save_messages(self) -> bool:
		"""Indicates whether messages generated by the reliability analysis will be saved
        If saved, messages will be part of the design point"""
		return interface.GetBoolValue(self._id, 'save_messages')

	@save_messages.setter
	def save_messages(self, value : bool):
		interface.SetBoolValue(self._id, 'save_messages', value)

	@property
	def reuse_calculations(self) -> bool:
		"""Indicates whether prior model results will be reused in the sensitivity analysis.

        This will speed up calculations when several analyses are performed, for which the same realizations
        will have to be executed, for example when calculating the sensitivity of several output parameters. But
        when a modification to the model is made, which is beyond the scope of the model definition, this
        leads to undesired results"""

		return interface.GetBoolValue(self._id, 'reuse_calculations')

	@reuse_calculations.setter
	def reuse_calculations(self, value : bool):
		interface.SetBoolValue(self._id, 'reuse_calculations', value)

	@property
	def sensitivity_method(self) -> SensitivityMethod:
		"""Defines the sensitivity algorithm"""
		return SensitivityMethod[interface.GetStringValue(self._id, 'sensitivity_method')]

	@sensitivity_method.setter
	def sensitivity_method(self, value : SensitivityMethod):
		interface.SetStringValue(self._id, 'sensitivity_method', str(value))

	@property
	def iterations(self) -> int:
		"""Number of iterations to apply"""
		return interface.GetIntValue(self._id, 'iterations')

	@iterations.setter
	def iterations(self, value : int):
		interface.SetIntValue(self._id, 'iterations', value)

	@property
	def low_value(self) -> float:
		"""The low value (defined as probability) a variable can be assigned to"""
		return interface.GetValue(self._id, 'low_value')

	@low_value.setter
	def low_value(self, value : float):
		interface.SetValue(self._id, 'low_value', value)

	@property
	def high_value(self) -> float:
		"""The high value (defined as probability) a variable can be assigned to"""
		return interface.GetValue(self._id, 'high_value')

	@high_value.setter
	def high_value(self, value : float):
		interface.SetValue(self._id, 'high_value', value)

	def is_valid(self) -> bool:
		"""Indicates whether the settings are valid"""
		return interface.GetBoolValue(self._id, 'is_valid')

	def validate(self):
		"""Prints the validity of the settings"""
		id_ = interface.GetIdValue(self._id, 'validate')
		if id_ > 0:
			validation_report = ValidationReport(id_)
			validation_report.print()

	def _set_variables(self, variables):
		new_stochast_settings = []
		for variable in variables:
			stochast_setting = self._stochast_settings[str(variable)]
			if stochast_setting is None:
				stochast_setting = StochastSettings(variable)
			new_stochast_settings.append(stochast_setting)
		self._stochast_settings = FrozenList(new_stochast_settings)
		interface.SetArrayIntValue(self._id, 'stochast_settings', [stochast_setting._id for stochast_setting in self._stochast_settings])

class SensitivityValue(FrozenObject):
	"""Contains the result of a sensitivity analysis for a specific input variable

    Several sensitivity values (one for each input variable) are listed in a sensitivity result, which
    contains the sensitivity results for a specific output parameter of a model."""

	def __init__(self, id, known_variables = None):
		self._id = id
		self._variable = None
		self._known_variables = known_variables
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['variable',
				'low',
				'medium',
				'high',
				'first_order_index',
				'total_index',
				'print']

	def __str__(self):
		return self.variable.name

	@property
	def variable(self) -> Stochast:
		"""The (input) stochastic variable to which these results apply"""
		if self._variable is None:
			variable_id = interface.GetIdValue(self._id, 'variable')
			if variable_id > 0:
				self._variable = self._get_variable_by_id(variable_id)

		return self._variable

	def _get_variable_by_id(self, variable_id) -> Stochast:
		if self._known_variables is not None:
			for variable in self._known_variables:
				if variable._id == variable_id:
					return variable
		return Stochast(variable_id)

	@property
	def low(self) -> float:
		"""The value the output parameter gets when the input variable is set to its low value"""
		return interface.GetValue(self._id, 'low')

	@property
	def medium(self) -> float:
		"""The value the output parameter gets when the input variable is set to its median value"""
		return interface.GetValue(self._id, 'medium')

	@property
	def high(self) -> float:
		"""The value the output parameter gets when the input variable is set to its high value"""
		return interface.GetValue(self._id, 'high')

	@property
	def first_order_index(self) -> float:
		"""The first order index of the output parameter due to variations of the input variable"""
		return interface.GetValue(self._id, 'first_order_index')

	@property
	def total_index(self) -> float:
		"""The total index of the output parameter due to variations of the input variable"""
		return interface.GetValue(self._id, 'total_index')

	def print(self, decimals=4):
		"""Prints the sensitivity value

        Parameters
        ----------
        decimals : int, optional
            The number of decimals to print"""

		self._print(0, decimals)

	def _print(self, indent = 0, decimals=4):
		pre = PrintUtils.get_space_from_indent(indent)
		if not isnan(self.low):
			print(pre + f'{self.variable.name}: low = {self.low:.{decimals}g}, medium = {self.medium:.{decimals}g}, high = {self.high:.{decimals}g}')
		elif not isnan(self.total_index):
			print(pre + f'{self.variable.name}: first order index = {self.first_order_index:.{decimals}g}, total index = {self.total_index:.{decimals}g}')


class SensitivityResult(FrozenObject):
	"""Contains the result of a sensitivity analysis for a specific output parameter

    A sensitivity result contains values for each input variable, listed in 'values'."""

	def __init__(self, id):
		self._id = id
		self._values = None
		self._messages = None
		self._realizations = None
		self._known_variables = None
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['identifier',
				'realizations',
				'messages',
				'print',
				'plot']

	def __str__(self):
		return self.identifier

	@property
	def identifier(self) -> str:
		"""Identification of the output parameter"""
		return interface.GetStringValue(self._id, 'identifier')

	@property
	def values(self) -> list[SensitivityValue]:
		"""List of sensitivity values per input variable. The values in this list contain the sensitivity results."""
		if self._values is None:
			sens_values = []
			sens_value_ids = interface.GetArrayIdValue(self._id, 'values')
			for sens_value_id in sens_value_ids:
				sens_values.append(SensitivityValue(sens_value_id, self._known_variables))
			self._values = FrozenList(sens_values)

		return self._values

	@property
	def realizations(self) -> list[Evaluation]:
		"""List of samples calculated by the reliability algorithm. Depends on the setting 'save_realizations' whether
        this list is provided"""
		if self._realizations is None:
			realizations = []
			realization_ids = interface.GetArrayIdValue(self._id, 'evaluations')
			for realization_id in realization_ids:
				realizations.append(Evaluation(realization_id))
			self._realizations = FrozenList(realizations)

		return self._realizations

	@property
	def messages(self) -> list[Message]:
		"""List of messages generated by the reliability algorithm. Depends on the setting 'save_messages' whether this
        list is provided"""
		if self._messages is None:
			messages = []
			message_ids = interface.GetArrayIdValue(self._id, 'messages')
			for message_id in message_ids:
				messages.append(Message(message_id))
			self._messages = FrozenList(messages)

		return self._messages

	def _set_variables(self, variables):
		self._known_variables = FrozenList(variables)

	def print(self, decimals=4):
		"""Prints the sensitivity result, including sensitivity values for each input variable.

        Parameters
        ----------
        decimals : int, optional.
            The number of decimals to print"""

		print('Parameter: ' + self.identifier)
		if len(self.values) > 0:
			print('Values:')
			for value in self.values:
				value._print(1, decimals)

	def plot(self):
		"""Shows a plot of the sensitivity in the form of a bar-chart"""
		self.get_plot().show()

	def get_plot(self) -> plt:
		"""Gets a plot object of the sensitivity in the form of a bar-chart"""

		plt.subplot()

		low_values = [value.low for value in self.values if not isnan(value.low)]
		medium_values = [value.medium for value in self.values if not isnan(value.medium)]
		high_values = [value.high for value in self.values if not isnan(value.high)]
		first_values = [value.first_order_index for value in self.values if not isnan(value.first_order_index)]
		total_values = [value.total_index for value in self.values if not isnan(value.total_index)]

		meaningful_values_count = sum(len(values) > 0 for values in [low_values, medium_values, high_values, first_values, total_values])

		bar_width = 1.0 / (meaningful_values_count + 1)

		x = np.arange(len(self.values))

		if len(low_values) > 0:
			plt.bar(x, low_values, color ='b', width = bar_width, label ='low')
			x = [x + bar_width for x in x]

		if len(medium_values) > 0:
			plt.bar(x, medium_values, color ='r', width = bar_width, label ='medium')
			x = [x + bar_width for x in x]

		if len(high_values) > 0:
			plt.bar(x, high_values, color ='g', width = bar_width, label ='high')
			x = [x + bar_width for x in x]

		if len(first_values) > 0:
			plt.bar(x, first_values, color ='y', width = bar_width, label ='first order')
			x = [x + bar_width for x in x]

		if len(total_values) > 0:
			plt.bar(x, total_values, color ='m', width = bar_width, label ='total')
			x = [x + bar_width for x in x]

		plt.xlabel('variables')
		plt.ylabel(self.identifier)

		increment = (meaningful_values_count - 1) / 2 * bar_width
		plt.xticks([r + increment for r in range(len(self.values))], [value.variable.name for value in self.values])

		plt.legend()

		return plt



# Copyright (C) Stichting Deltares. All rights reserved.
#
# This file is part of the Probabilistic Library.
#
# The Probabilistic Library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# All names, logos, and references to "Deltares" are registered trademarks of
# Stichting Deltares and remain full property of Stichting Deltares at all times.
# All rights reserved.
#


from __future__ import annotations
import sys
from multiprocessing import Pool, cpu_count
from typing import FrozenSet
from types import FunctionType, MethodType
from enum import Enum

from .statistic import Stochast, DistributionType, CorrelationMatrix, CopulaCorrelation, SelfCorrelationMatrix, Scenario, CorrelationType
from .reliability import DesignPoint, ReliabilityMethod, Settings, CombineSettings, ExcludingCombineSettings, LimitStateFunction
from .sensitivity import SensitivityResult, SensitivityValue, SensitivitySettings, SensitivityMethod
from .uncertainty import UncertaintyResult, UncertaintySettings, UncertaintyMethod
from .logging import Evaluation, Message, ValidationReport
from .utils import FrozenObject, FrozenList, CallbackList
from . import interface

import inspect

if not interface.IsLibraryLoaded():
	interface.LoadDefaultLibrary()

class ZModel(FrozenObject):
	"""Wrapper around a python function or method.

    This class is created by a `ModelProject` when its model is set.

    The ZModel provides additional information about its input and output parameters.
    This is based on reflection when the ZModel is created."""

	_callback = None
	_multiple_callback = None

	def __init__(self, callback = None, output_parameter_size = 1):
		self._model = None
		source_code = None
		if isinstance(callback, str):
			source_code = callback
			try:
				dynamic_code = compile(callback, '<string>', 'exec')
				code_object = dynamic_code.co_consts[0]
				callback = FunctionType(code_object, globals(), code_object.co_name)
			except Exception:
				callback = source_code # so that _is_function will be False
		self._is_function = inspect.isfunction(callback) or inspect.ismethod(callback)
		self._is_dirty = False
		self._has_arrays = False
		self._array_sizes = None
		self._pool = None
		self._max_processes = 1
		self._project = None
		self._project_id = 0
		self._z_values_size = 0
		ZModel._index = 0;
		ZModel._callback = callback

		if self._is_function:
			self._model_name = callback.__name__
			self._input_parameters = self._get_input_parameters(callback)
			if source_code != None:
				self._output_parameters = self._get_output_parameters(source_code, output_parameter_size)
			else:
				self._output_parameters = self._get_output_parameters(callback, output_parameter_size)
		else:
			self._model_name = ''
			self._input_parameters = []
			self._output_parameters = []

		super()._freeze()

	def __dir__(self):
		return ['name',
				'input_parameters',
				'output_parameters',
 				'print']

	def __del__(self):
		try:
			if not self._pool is None:
				self._pool.close()
		except Exception as err:
			print(f"Unexpected {err=}, {type(err)=}")

	def __str__(self):
		return self.name

	@property
	def name(self) -> str:
		"""Name of the model"""
		return self._model_name

	def _get_input_parameters(self, function):
		sig = inspect.signature(function)
		parameters = []
		index = 0
		for name, param in sig.parameters.items():
			modelParameter = ModelParameter()
			modelParameter.name = name
			if param.default != param.empty:
				modelParameter.default_value = param.default
			if param.annotation != param.empty:
				modelParameter.is_array = '[' in str(param)
			modelParameter.index = index
			parameters.append(modelParameter)
			index += 1

		return FrozenList(parameters)

	def _get_output_parameters(self, function, output_parameter_size = 1):
		parameters = []
		if isinstance(function, str):
			source = function
		else:
			source = inspect.getsource(function)
		lines = source.splitlines()
		for line in lines:
			if line.strip().startswith('return'):
				line = line.replace('return', '')
				line = line.strip().split('#')[0]
				line = line.lstrip('[').rstrip(';').rstrip(']')
				words = line.split(',')
				parameters = [word.strip() for word in words]

		for i in range(len(parameters)):
			modelParameter = ModelParameter()
			modelParameter.name = parameters[i]
			modelParameter.index = i
			parameters[i] = modelParameter
			if (len(parameters) == 1 and output_parameter_size > 1):
				parameters[i].is_array = True
				parameters[i].array_size = output_parameter_size

		return FrozenList(parameters)

	def is_model_valid(self) -> bool:
		"""Indicates whether the model is valid"""
		if self._is_function:
			return True
		elif not self._model is None:
			return self._model.is_model_valid()
		else:
			return False

	@property
	def input_parameters(self) -> list[ModelParameter]:
		"""List of input parameters"""
		return self._input_parameters

	@property
	def output_parameters(self) -> list[ModelParameter]:
		"""List of output parameters"""
		return self._output_parameters

	def _set_callback(self, callback):
		ZModel._callback = callback

	def _set_multiple_callback(self, multiple_callback):
		ZModel._multiple_callback = multiple_callback

	def _set_model(self, value):
		self._model = value

	def set_max_processes(self, value : int):
		"""Sets the maximum number of parallel processes"""
		self._max_processes = value
		if not self._model is None:
			self._model.set_max_processes(value)

	def initialize_for_run(self):
		"""Method to be called before the first run (this method is used internally by `ModelProject`)"""
		self._has_arrays = False
		for parameter in self.input_parameters:
			if parameter.is_array:
				self._has_arrays = True

		if self._has_arrays:
			self._array_sizes = []
			for parameter in self.input_parameters:
				if parameter.is_array:
					self._array_sizes.append(parameter.array_size)
				else:
					self._array_sizes.append(-1)

		self._z_values_size = 0
		for parameter in self.output_parameters:
			if parameter.is_array:
				self._z_values_size += parameter.array_size
			else:
				self._z_values_size += 1

		if not self._model is None:
			self._model.initialize_for_run()

		if self._is_function:
			if self._max_processes > 1:
				self._pool = Pool(self._max_processes)
			elif self._max_processes < 1:
				self._pool = Pool()
			else:
				self._pool = None

	def update(self):
		"""Updates input and output parameters (this method is used internally by `ModelProject`)"""
		if not self._model is None:
			if self._model.is_dirty():
				self._model.update_model()
				return True

		return False

	def _get_args(self, values):
		args = []
		index = 0;
		for i in range(len(self._array_sizes)):
			if self._array_sizes[i] == -1:
				args.append(values[index])
				index += 1
			else:
				arg_array = []
				for j in range(self._array_sizes[i]):
					arg_array.append(values[index])
					index += 1
				args.append(arg_array)
		return args

	def run_multiple(self, samples):
		"""Performs the execution of multiple samples by the model (used internally)"""
		if self._is_function and self._pool is None:
			for sample in samples:
				self.run(sample)
		elif self._is_function and not self._pool is None:
			results = {}
			for sample in samples:
				sample_input = self._get_input(sample)
				results[sample] = self._pool.apply_async(func=ZModel._callback, args=(*sample_input,))
			for sample in samples:
				z = results[sample].get()
				self._assign_output(sample, z)
		else:
			ZModel._multiple_callback(samples)

	def run(self, sample):
		"""Performs the execution of a sample by the model (used internally)"""
		if self._is_function:
			sample_input = self._get_input(sample)
			z = ZModel._callback(*sample_input)
			self._assign_output(sample, z)
		else:
			z = ZModel._callback(sample);

	def _get_input(self, sample):
		if self._has_arrays:
			return self._get_args(sample.input_values)
		else:
			return sample.input_values

	def _assign_output(self, sample, z):
		if type(z) is list or type(z) is tuple:
			for i in range(self._z_values_size):
				sample.output_values[i] = z[i]
		else:
			sample.output_values[0] = z

	def _run_callback(sample_input):
		return ZModel._callback(*sample_input)

	def print(self):
		"""Prints the model with all input and output parameters"""
		pre = '  '
		if not self.name == '':
			print(f'Model {self.name}:')
		print('Input parameters:')
		for input_parameter in self.input_parameters:
			if input_parameter.is_array:
				print(pre + f'{input_parameter.name}[{input_parameter.array_size}]')
			else:
				print(pre + f'{input_parameter.name}')
		print('Output parameters:')
		for output_parameter in self.output_parameters:
			if output_parameter.is_array:
				print(pre + f'{output_parameter.name}[{output_parameter.array_size}]')
			else:
				print(pre + f'{output_parameter.name}')

class ZModelContainer:
	"""Wrapper of a `ZModel`, used internally by ptk.whl to assign a PTK model to a `ModelProject`"""

	def get_model(self) -> ZModel|None:
		"""Gets a ZModel"""
		return None

	def is_model_valid(self) -> bool:
		"""Indicates whether the model is valid"""
		return True

	def is_dirty(self):
		return False

	def update_model(self):
		pass

class ModelParameter(FrozenObject):
	"""Input or output parameter of a model

    A model parameter is part of a `ZModel`, as one of its input or output parameters. It is generated when a
    `ZModel` is created. By reflection and type hints as much as possible information about the parameter is
    provided."""

	def __init__(self, id = None):
		if id is None:
			self._id = interface.Create('model_parameter')
		else:
			self._id = id
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['name',
				'index',
				'default_value',
				'is_array',
				'array_size']

	@property
	def name(self) -> str:
		"""Name of the parameter"""
		return interface.GetStringValue(self._id, 'name')

	@name.setter
	def name(self, value):
		interface.SetStringValue(self._id, 'name', value)

	@property
	def index(self) -> int:
		"""Sequence number of the parameter in the list of input or output parameters of a `ZModel`"""
		return interface.GetIntValue(self._id, 'index')

	@index.setter
	def index(self, value : int):
		interface.SetIntValue(self._id, 'index', value)

	@property
	def default_value(self) -> float:
		"""Default value of the parameter"""
		return interface.GetValue(self._id, 'default_value')

	@default_value.setter
	def default_value(self, value : float):
		interface.SetValue(self._id, 'default_value', value)

	@property
	def is_array(self) -> bool:
		"""Indicates whether the parameter is an array"""
		return interface.GetBoolValue(self._id, 'is_array')

	@is_array.setter
	def is_array(self, value : bool):
		interface.SetBoolValue(self._id, 'is_array', value)

	@property
	def array_size(self) -> int:
		"""Array size, in case the parameter is an array"""
		return interface.GetIntValue(self._id, 'array_size')

	@array_size.setter
	def array_size(self, value : int):
		interface.SetIntValue(self._id, 'array_size', value)

	def __str__(self):
		return self.name

class ModelProject(FrozenObject):
	"""Base class for projects, which contain a model

    When a model is set to `model`, the model input parameters and output parameters are updated. Based on
    these parameters, variables, correlation matrix and settings and settings per variable are generated in
    this class."""

	_project_id = 0
	_zmodel = None

	def __init__(self):

		self._known_variables = []
		self._variables = FrozenList()
		self._correlation_matrix = CorrelationMatrix()
		self._copulas = None
		self._output_parameters = FrozenList()
		self._settings = None
		self._model = None
		self._correlation_type = CorrelationType.gaussian
		# do not freeze, this will be done by inheritors

	def _initialize_callbacks(self, project_id):

		ModelProject._project_id = project_id
		self._project_id = project_id
		self._callback = interface.CALLBACK(self._performCallBack)
		self._multiple_callback = interface.MULTIPLE_CALLBACK(self._perform_multiple_callback)

		interface.SetCallBack(project_id, 'model', self._callback)
		interface.SetMultipleCallBack(project_id, 'model', self._multiple_callback)
		interface.SetBoolValue(project_id, 'callback_assigned', False)

	def _set_settings(self, settings):
		self._settings = settings

	@interface.CALLBACK
	def _performCallBack(values, size, output_values):
		sample = _Sample(values[:size], output_values)
		ModelProject._zmodel.run(sample)

	@interface.MULTIPLE_CALLBACK
	def _perform_multiple_callback(sample_count, values, input_size, output_values):
		samples = []
		for i in range(sample_count):
			samples.append(_Sample(values[i][:input_size], output_values[i]))
		ModelProject._zmodel.run_multiple(samples)

	def is_valid(self) -> bool:
		"""Indicates whether the settings are valid"""
		self._update()
		return interface.GetBoolValue(self._id, 'is_valid')

	def validate(self):
		"""Prints the validity of the settings"""
		self._update()
		id_ = interface.GetIdValue(self._id, 'validate')
		if id_ > 0:
			validation_report = ValidationReport(id_)
			validation_report.print()

	@property
	def variables(self) -> list[Stochast]:
		"""List of variables based on the input parameters of the model"""
		self._check_model()
		return self._variables

	@property
	def correlation_type(self) -> CorrelationType:
		return self._correlation_type

	@correlation_type.setter
	def correlation_type(self, value):
		# if the type changes, create an new correlation class for the type now in use, and make the other inaccessable
		if (value != self._correlation_type):
			if (value == CorrelationType.gaussian):
				self._correlation_matrix = CorrelationMatrix()
				self._correlation_matrix._set_variables(self._copulas._variables)
				self._copulas = None
			else:
				self._copulas = CopulaCorrelation()
				self._copulas._set_variables(self._correlation_matrix._variables)
				self._correlation_matrix = None
			self._correlation_type = value

	@property
	def correlation_matrix(self) -> CorrelationMatrix:
		"""Correlation matrix based on the input parameters of the model"""
		self._check_model()
		return self._correlation_matrix

	@property
	def copulas(self) -> CopulaCorrelation:
		"""Correlation copula based on the input parameters of the model"""
		self._check_model()
		return self._copulas

	@property
	def model(self) -> ZModel:
		"""Method which serves as a model. A model is a function which calculates real world
        results based on real world input data (or is an academic function). It often relates
        to physical processes and is deterministic (it does not use uncertainty)

        When a model is set, it accepts a python function or python class method. Alternatively,
        a string defining a function is accepted too. The model should accept a number of input
        values (floats) or array of input values and returns a single value (float), an array of
        floats or a tuple of floats.

        When set, the function/method/string is wrapped in a `ZModel`. The ZModel has information
        about its input and output parameters (derived from the function signature). When the model
        is retrieved, the ZModel object is returned.

        It is also possible to set a `ZModelContainer`, which can provide a `ZModel`. PTK models
        are set with a `ZModelContainer` (using ptk.whl)"""

		if not self._model is None:
			self._model._project = self
		return self._model

	@model.setter
	def model(self, value : FunctionType | MethodType | str | ZModel | ZModelContainer):
		if isinstance(value, tuple):
			output_parameter_size = value[1]
			value = value[0]
		else:
			output_parameter_size = 1

		if inspect.isfunction(value) or inspect.ismethod(value) or isinstance(value, str):
			self._model = ZModel(value, output_parameter_size)
			self._update_model()
		elif isinstance(value, ZModel):
			self._model = value
			self._share(value._project)
			self._update_model()
		elif isinstance(value, ZModelContainer):
			self._model = value.get_model()
		else:
			raise ValueError('ZModel container expected')

		interface.SetBoolValue(self._project_id, 'callback_assigned', self._model.is_model_valid())

	def _check_model(self):
		if not self._model is None:
			if self._model.update():
				self._update_model()

	def _share(self, shared_project):
		id1 = self._id
		id2 = shared_project._id
		self._known_variables.extend(shared_project.variables)
		interface.SetIntValue(self._id, 'share_project', shared_project._id)

	def _update_model(self):
		interface.SetArrayIntValue(self._project_id, 'input_parameters', [input_parameter._id for input_parameter in self._model.input_parameters])
		interface.SetArrayIntValue(self._project_id, 'output_parameters', [output_parameter._id for output_parameter in self._model.output_parameters])
		interface.SetStringValue(self._project_id, 'model_name', self._model.name)

		variables = []
		variable_ids = interface.GetArrayIdValue(self._project_id, 'stochasts')
		for variable_id in variable_ids:
			variable = None
			for known_variable in self._known_variables:
				if known_variable._id == variable_id:
					variable = known_variable
			if variable is None:
				variable = Stochast(variable_id)
				self._known_variables.append(variable)
			variables.append(variable)
		self._variables = FrozenList(variables)

		if (self._correlation_type == CorrelationType.gaussian):
			self._correlation_matrix._set_variables(variables)
		else:
			self._copulas._set_variables(variables)
		self._settings._set_variables(variables)
		for var in self._variables:
			var._set_variables(self._variables)

		self._output_parameters = self._model.output_parameters

	def _update(self):
		self._check_model()

		if (self._correlation_type == CorrelationType.gaussian):
			interface.SetIntValue(self._project_id, 'correlation_matrix', self._correlation_matrix._id)
		else:
			interface.SetIntValue(self._project_id, 'copula_correlation', self._copulas._id)
		interface.SetIntValue(self._project_id, 'settings', self._settings._id)
		if hasattr(self.settings, 'stochast_settings'):
			interface.SetArrayIntValue(self.settings._id, 'stochast_settings', [stochast_setting._id for stochast_setting in self.settings.stochast_settings])
		if self._model != None:
			if hasattr(self.settings, 'max_parallel_processes'):
				self._model.set_max_processes(self.settings.max_parallel_processes)
			self._model.initialize_for_run()
		ModelProject._zmodel = self._model

	def _run(self):
		if (self.is_valid()):
			interface.Execute(self._project_id, 'run')
		else:
			# print the validation messages
			self.validate()

class RunValuesType(Enum):
	"""Enumeration which defines which value to use from a stochastic variable"""
	median_values = 'median_values'
	mean_values = 'mean_values'
	design_values = 'design_values'
	def __str__(self):
		return str(self.value)

class _Sample(FrozenObject):
	"""Defines a sample, used internally"""
	def __init__(self, input_values, output_values):
		self.input_values = input_values
		self.output_values = output_values
		super()._freeze()

class RunProjectSettings(FrozenObject):
	"""Settings for a model run"""

	def __init__(self):
		self._id = interface.Create('run_project_settings')
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['run_values_type',
		        'reuse_calculations',
		        'validate',
		        'is_valid']

	@property
	def run_values_type(self) -> RunValuesType:
		"""Defines which value to extract from the stochastic variables"""
		return RunValuesType[interface.GetStringValue(self._id, 'run_values_type')]

	@run_values_type.setter
	def run_values_type(self, value : RunValuesType):
		interface.SetStringValue(self._id, 'run_values_type', str(value))

	@property
	def reuse_calculations(self) -> bool:
		"""Indicates whether previous model results will be reused by the model run."""
		return interface.GetBoolValue(self._id, 'reuse_calculations')

	@reuse_calculations.setter
	def reuse_calculations(self, value : bool):
		interface.SetBoolValue(self._id, 'reuse_calculations', value)

	def is_valid(self) -> bool:
		"""Indicates whether the settings are valid"""
		return interface.GetBoolValue(self._id, 'is_valid')

	def validate(self):
		"""Prints the validity of the settings"""
		id_ = interface.GetIdValue(self._id, 'validate')
		if id_ > 0:
			validation_report = ValidationReport(id_)
			validation_report.print()

	def _set_variables(self, variables):
		pass

class RunProject(ModelProject):
	"""Project for a running a model. This is the main entry point for running a model.

    This class is based on the `ModelProject` class. The model to use is defined in
    this class in `ModelProject.model`. When a model is set, variables and settings
    are generated.

    To run the model, use the `run` method. This results are stored in `realization`.

    """

	def __init__(self):
		super().__init__()
		self._id = interface.Create('run_project')
		self._realization = None
		self._initialize_callbacks(self._id)
		self._set_settings(RunProjectSettings())
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['variables',
				'correlation_matrix',
				'settings',
				'model',
				'run',
				'realization',
				'validate',
				'is_valid']

	@property
	def settings(self) -> RunProjectSettings:
		"""Settings for running a model"""
		self._check_model()
		return self._settings

	def run(self):
		"""Performs the model run and puts the result in `realization`"""
		self._realization = None
		self._run()

	@property
	def realization(self) -> Evaluation:
		"""Realization of the performed model run"""
		if self._realization is None:
			realizationId = interface.GetIdValue(self._id, 'realization')
			if realizationId > 0:
				self._realization = Evaluation(realizationId)

		return self._realization

class SensitivityProject(ModelProject):
	"""Project for a sensitivity analysis. This is the main entry point for performing a sensitivity analysis.

    This class is based on the `ModelProject` class. The model to use is defined in this class in `ModelProject.model`.
    When a model is set, variables, correlation matrix and settings and settings per variable are generated.

    To run a sensitivity analysis, use the `run` method. This results are stored in a list of `results`, where each
    result corresponds with an output parameter. A shortcut is provided in `result`, which corresponds with the
    first output parameter."""

	def __init__(self):
		super().__init__()
		self._id = interface.Create('sensitivity_project')

		self._result = None
		self._results = None

		self._initialize_callbacks(self._id)
		self._set_settings(SensitivitySettings())
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['variables',
				'correlation_matrix',
				'settings',
				'model',
				'parameter',
				'run',
				'result',
				'results',
				'validate',
				'is_valid',
				'total_model_runs']

	@property
	def parameter(self) -> str:
		"""Output parameter for which the sensitivity analysis should be performed, if left blank it will
        be performed for all output parameters"""
		return interface.GetStringValue(self._id, 'parameter')

	@parameter.setter
	def parameter(self, value : str | ModelParameter):
		interface.SetStringValue(self._id, 'parameter', str(value))

	@property
	def settings(self) -> SensitivitySettings :
		"""Settings for the sensitivity algorithm"""
		self._check_model()
		return self._settings

	def run(self):
		"""Performs the sensitivity analysis
        Results will be stored in `results`, for the first output `parameter` in `result`."""
		self._result = None
		self._results = None

		self._run()

	@property
	def result(self) -> SensitivityResult:
		"""Sensitivity result corresponding with the output parameters"""
		if self._result is None:
			result_id = interface.GetIdValue(self._id, 'result')
			if result_id > 0:
				self._result = SensitivityResult(result_id)

		return self._result

	@property
	def results(self) -> list[SensitivityResult]:
		"""List of sensitivity results, corresponding with the output parameters"""
		if self._results is None:
			results = []
			result_ids = interface.GetArrayIdValue(self._id, 'results')
			for result_id in result_ids:
				result = SensitivityResult(result_id)
				result._set_variables(self.variables)
				results.append(result)
			self._results = FrozenList(results)

		return self._results

	@property
	def total_model_runs(self) -> int:
		"""Total model runs performed by the last `run`"""
		return interface.GetIntValue(self._id, 'total_model_runs')

class UncertaintyProject(ModelProject):
	"""Project for an uncertainty analysis. This is the main entry point for performing an uncertainty analysis.

    This class is based on the `ModelProject` class. The model to use is defined in this class in 'ModelProject.model'.
    When a model is set, variables, correlation matrix and settings and settings per variable are generated.

    The uncertainty analysis calculates the uncertainty for the output parameter selected in 'parameter'. If left blank,
    uncertainty analyses are performed for each output parameter.

    To run an uncertainty analysis, use the `run` method. This results in a list of `results`. Each result contains the
    uncertainty for a certain output parameter. Part of the result is a stochast, which contains the distribution
    of the output parameter. A shortcut to the stochasts is available via `stochasts`. The first output parameter
    result can also be accessed by `result` and `stochast`.

    If correlation between output parameters is requested via the `settings`, the resulting correlation matrix can be found
    in `output_correlation_matrix`."""

	def __init__(self):
		super().__init__()
		self._id = interface.Create('uncertainty_project')

		self._stochast = None
		self._stochasts = None
		self._result = None
		self._results = None
		self._output_correlation_matrix = None

		self._initialize_callbacks(self._id)
		self._set_settings(UncertaintySettings())
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['variables',
				'correlation_matrix',
				'settings',
				'model',
				'parameter',
				'run',
				'stochast',
				'stochasts',
				'result',
				'results',
				'output_correlation_matrix',
				'validate',
				'is_valid',
				'total_model_runs']

	@property
	def parameter(self) -> str:
		"""Output parameter for which the uncertainty analysis should be performed, if left blank it will
        be performed for all output parameters"""
		return interface.GetStringValue(self._id, 'parameter')

	@parameter.setter
	def parameter(self, value : str | ModelParameter):
		interface.SetStringValue(self._id, 'parameter', str(value))

	@property
	def settings(self) -> UncertaintySettings :
		"""Settings for the uncertainty algorithm"""
		self._check_model()
		return self._settings

	def run(self):
		"""Performs the uncertainty analysis
        Results will be stored in `results`, for the first output `parameter` in `result`. The stochasts in
        the results are available directly in `stochasts` and `stochast`"""

		self._stochast = None
		self._stochasts = None
		self._result = None
		self._results = None
		self._output_correlation_matrix = None

		self._run()

	@property
	def stochast(self) -> Stochast:
		"""Stochast distribution of the first output `parameter`"""
		if self._stochast is None:
			if not self.result is None:
				self._stochast = self.result.variable
		return self._stochast

	@property
	def stochasts(self) -> list[Stochast]:
		"""List of probability distributions, corresponding with the output parameters"""
		if self._stochasts is None:
			stochasts = []
			for result in self.results:
				if not result is None:
					stochasts.append(result.variable)
				else:
					stochasts.append(None)
			self._stochasts = FrozenList(stochasts)

		return self._stochasts

	@property
	def result(self) -> UncertaintyResult:
		"""Uncertainty result of the first output parameter"""
		if self._result is None:
			result_id = interface.GetIdValue(self._id, 'uncertainty_result')
			if result_id > 0:
				self._result = UncertaintyResult(result_id, self.variables)

		return self._result

	@property
	def results(self) -> list[UncertaintyResult]:
		"""List of uncertainty results, corresponding with the output parameters"""
		if self._results is None:
			results = []
			result_ids = interface.GetArrayIdValue(self._id, 'uncertainty_results')
			for result_id in result_ids:
				results.append(UncertaintyResult(result_id, self.variables))
			self._results = FrozenList(results)

		return self._results

	@property
	def output_correlation_matrix(self) -> CorrelationMatrix:
		"""Correlation matrix between output parameters and possibly with input parameters
        Depends on settings `UncertaintySettings.calculate_correlations` and
        `UncertaintySettings.calculate_input_correlations` whether this correlation matrix is generated"""

		if self._output_correlation_matrix is None:
			correlationMatrixId = interface.GetIdValue(self._id, 'output_correlation_matrix')
			if correlationMatrixId > 0:
				self._output_correlation_matrix = CorrelationMatrix(correlationMatrixId)
				self._output_correlation_matrix._update_variables(self.variables.get_list() + self.stochasts.get_list())

		return self._output_correlation_matrix

	@property
	def total_model_runs(self) -> int:
		"""Total model runs performed by the last `run`"""
		return interface.GetIntValue(self._id, 'total_model_runs')

class ReliabilityProject(ModelProject):
	"""Project for reliability analysis. This is the main entry point for performing a reliability analysis.

    This class is based on the `ModelProject` class. The model to use is defined in this class in `ModelProject.model`.
    When a model is set, variables, correlation matrix and settings and settings per variable are generated.

    To run a reliability analysis, use the `run` method. This results in a `design_point`, where the reliability
    index, alpha values and, if specified in the settings, an overview of performed realizations and generated messages."""

	def __init__(self):
		super().__init__()
		self._id = interface.Create('project')

		self._limit_state_function = None
		self._design_point = None
		self._fragility_curve = None
		self._initialized = False

		self._initialize_callbacks(self._id)
		self._set_settings(Settings())
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['variables',
				'correlation_matrix',
				'limit_state_function',
				'settings',
				'model',
				'run',
				'design_point',
				'validate',
				'is_valid',
				'total_model_runs']

	@property
	def settings(self) -> Settings:
		"""Settings for the algorithm performing the reliability analysis"""
		self._check_model()
		return self._settings

	@property
	def limit_state_function(self) -> LimitStateFunction:
		"""Defines the transformation of the model output to a z-value
        By default, the first value in the model output is used as z-value"""

		if self._limit_state_function is None:
			lsf_id = interface.GetIdValue(self._id, 'limit_state_function')
			if lsf_id > 0:
				self._limit_state_function = LimitStateFunction(lsf_id)
		return self._limit_state_function

	def run(self):
		"""Performs the reliability analysis
        Results in a `design_point`, which is part of this project. When failed, the `design_point` is empty."""

		self._design_point = None
		self._fragility_curve = None
		self._initialized = False
		self._run()

	@property
	def design_point(self) -> DesignPoint:
		"""The resulting design point of a reliability analysis, invoked by `run`
        Is empty when the run failed"""
		if self._design_point is None:
			designPointId = interface.GetIdValue(self._id, 'design_point')
			if designPointId > 0:
				self._design_point = DesignPoint(designPointId, self._get_variables())

		return self._design_point

	def _get_variables(self):
		variables = []
		variables.extend(self.variables)
		for variable in variables:
			for array_variable in variable.array_variables:
				if array_variable not in variables:
					variables.append(array_variable)
		return variables

	@property
	def total_model_runs(self) -> int:
		"""Total model runs performed by the last `run`"""
		return interface.GetIntValue(self._id, 'total_model_runs')

class CombineProject(FrozenObject):
	"""Project for combining design points. This is the main entry point for performing combining design points.

    The design points to be combined should be added to the list of `design_points`.

    To run the combination, use the `run` method. This results in a `design_point`, where the
    reliability index reflects the combined reliability index. The original `design_points` are
    added to the `probabilistic_library.reliability.DesignPoint.contributing_design_points` of the
    resulting design point.
    """

	def __init__(self):
		self._id = interface.Create('combine_project')

		self._design_points = CallbackList(self._design_points_changed)
		self._settings = CombineSettings()
		self._correlation_matrix = SelfCorrelationMatrix()
		self._design_point_correlation_matrix = CorrelationMatrix()
		self._design_point = None
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['design_points',
				'settings',
				'correlation_matrix',
				'run',
				'design_point_correlation_matrix',
				'design_point']

	def _design_points_changed(self):
		variables = []
		for design_point in self._design_points:
			variables.extend(design_point.get_variables())
		self._correlation_matrix._set_variables(variables)
		self._design_point_correlation_matrix._set_variables(variables)

	@property
	def design_points(self) -> list[DesignPoint]:
		"""List of design points to be combined"""
		return self._design_points

	@property
	def settings(self) -> CombineSettings:
		"""Settings for the combination algorithm"""
		return self._settings

	@property
	def correlation_matrix(self) -> SelfCorrelationMatrix:
		"""Auto correlation matrix, holds correlations between same named variables in the `design_points`"""
		return self._correlation_matrix

	def _update(self):
		interface.SetArrayIntValue(self._id, 'design_points', [design_point._id for design_point in self._design_points])
		interface.SetIntValue(self._id, 'settings', self._settings._id)
		interface.SetIntValue(self._id, 'correlation_matrix', self._correlation_matrix._id)
		interface.SetIntValue(self._id, 'design_point_correlation_matrix', self._design_point_correlation_matrix._id)

	def is_valid(self) -> bool:
		"""Indicates whether the combine project is valid"""
		self._update()
		return interface.GetBoolValue(self._id, 'is_valid')

	def validate(self):
		"""Prints the validity of the combine project"""
		self._update()
		id_ = interface.GetIdValue(self._id, 'validate')
		if id_ > 0:
			validation_report = ValidationReport(id_)
			validation_report.print()

	def run(self):
		"""Performs the combination of the design point.
        Results in a `design_point`, which is part of this project. When failed, the `design_point` is empty."""
		self._design_point = None
		# update performed by is_valid
		if (self.is_valid()):
			interface.Execute(self._id, 'run')
		else:
			print('run not executed, input is not valid')

	@property
	def design_point(self) -> DesignPoint:
		"""The resulting combined design point, invoked by `run`
        Is empty when the run failed. The original `design_points` are added to the
        `probabilistic_library.reliability.DesignPoint.contributing_design_point`
        """
		if self._design_point is None:
			designPointId = interface.GetIdValue(self._id, 'design_point')
			if designPointId > 0:
				variables = []
				for design_point in self._design_points:
					variables.extend(design_point.get_variables())
				self._design_point = DesignPoint(designPointId, variables, self._design_points)
		return self._design_point

class ExcludingCombineProject(FrozenObject):
	"""Project for combining design points exclusively or, otherwise stated, design points calculated for a
    scenario. This is the main entry point for this operation.

    Excluding design points refer to design points generated for a scenario. Each scenario has a
    probability too, but scenarios are mutually exclusive. Probabilities of scenarios should add up to 1.

    The design points to be combined should be added to the list of `design_points`. The list of
    `scenarios` should correspond with the list of `design_points`.

    To run the combination, use the `run` method. This results in a `design_point`, where the
    reliability index reflects the combined reliability index. The original `design_points` are
    added to the `probabilistic_library.reliability.DesignPoint.contributing_design_points` of the resulting design point.

    """

	def __init__(self):
		self._id = interface.Create('excluding_combine_project')

		self._design_points = CallbackList(self._design_points_changed)
		self._scenarios = CallbackList(self._scenarios_changed)
		self._settings = ExcludingCombineSettings()
		self._design_point = None
		self._synchronizing = False
		self._dirty = True

		interface.SetIntValue(self._id, 'settings', self._settings._id)
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['design_points',
				'scenarios',
				'settings',
				'is_valid',
				'validate',
				'run',
				'design_point']

	def _design_points_changed(self):
		if not self._synchronizing:
			self._dirty = True

	def _scenarios_changed(self):
		if not self._synchronizing:
			# replace floats by Scenario
			self._synchronizing = True
			for i in range(len(self._scenarios)):
				if isinstance(self._scenarios[i], int) or isinstance(self._scenarios[i], float):
					val = self._scenarios[i]
					self._scenarios[i] = Scenario()
					self._scenarios[i].probability = val
			self._synchronizing = False
			self._dirty = True

	def _update(self):
		if self._dirty:
			interface.SetArrayIntValue(self._id, 'design_points', [design_point._id for design_point in self._design_points])
			interface.SetArrayIntValue(self._id, 'scenarios', [scenario._id for scenario in self._scenarios])
			self._dirty = False

	@property
	def design_points(self) -> list[DesignPoint]:
		"""List of design points to be combined"""
		return self._design_points

	@property
	def scenarios(self) -> list[Scenario]:
		"""List of scenarios corresponding with design points to be combined"""
		return self._scenarios

	@property
	def settings(self) -> ExcludingCombineSettings:
		"""Settings for the combine algorithm"""
		return self._settings

	def is_valid(self) -> bool:
		"""Indicates whether the excluding combine project is valid"""
		self._update()
		return interface.GetBoolValue(self._id, 'is_valid')

	def validate(self):
		"""Prints the validity of the excluding combine project"""
		self._update()
		id_ = interface.GetIdValue(self._id, 'validate')
		if id_ > 0:
			validation_report = ValidationReport(id_)
			validation_report.print()

	def run(self):
		"""Performs the excluding combination of the design point.
        Results in a design point, which is part of this project. When failed, the `design_point` is empty."""
		self._update()
		self._design_point = None
		if (self.is_valid()):
			interface.Execute(self._id, 'run')
		else:
			# print validation messages
			self.validate()

	@property
	def design_point(self):
		"""The resulting excluding combined design point, invoked by `run`.
        Is empty when the run failed. The original `design_points` are added to the `contributing_design_points`"""
		if self._design_point is None:
			design_point_id = interface.GetIdValue(self._id, 'design_point')
			if design_point_id > 0:
				variables = []
				for design_point in self._design_points:
					variables.extend(design_point.get_variables())
				self._design_point = DesignPoint(design_point_id, variables, self._design_points)
		return self._design_point

class LengthEffectProject(FrozenObject):
	"""Project for applying the length effect to a design point, also known as upscaling in space. This is
    the main entry point for applying the length effect.

    When a design point is valid for a certain section or cross section, it can be useful to make it
    applicable to a longer section. This operation takes into account the length effect or upscaling.
    Each input variable is valid for a certain length. These lengths are stored in the `correlation_lengths`
    list. The length effect applied to the design point results in a design point valid for the requested `length`.

    To run the combination, use the `LengthEffectProject.run` method. This results in a `design_point`, where the
    reliability index reflects the length effect applied reliability index. The original
    `LengthEffectProject.design_point_cross_section` is added to the
    `probabilistic_library.reliability.DesignPoint.contributing_design_points`  of the resulting design point.

    ```mermaid
    classDiagram
        class LengthEffectProject{
            +design_point_cross_section DesignPoint
            +design_point DesignPoint
            +run()
        }

        class Stochast{}
        class SelfCorrelationMatrix{}
        class Stochast{}

        DesignPoint "input" <-- LengthEffectProject
        DesignPoint "result" <-- LengthEffectProject
        SelfCorrelationMatrix <-- LengthEffectProject
        Stochast "*" <-- SelfCorrelationMatrix
    ```
    """

	def __init__(self):
		self._id = interface.Create('length_effect_project')
		self._design_point_cross_section = DesignPoint()
		self._correlation_matrix = SelfCorrelationMatrix()
		self._design_point = None
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['design_point_cross_section',
				'correlation_lengths'
				'length',
				'correlation_matrix',
				'run',
				'design_point']

	@property
	def design_point_cross_section(self) -> DesignPoint:
		"""The design point to which the length effect will be applied"""
		return self._design_point_cross_section

	@design_point_cross_section.setter
	def design_point_cross_section(self, value: DesignPoint):
		self._design_point_cross_section = value
		variables = value.get_variables()
		self._correlation_matrix._set_variables(variables)

	@property
	def correlation_matrix(self) -> SelfCorrelationMatrix:
		"""Auto correlation matrix, holds correlations of variables in the `design_point_cross_section`"""
		return self._correlation_matrix

	@property
	def correlation_lengths(self) -> list[float]:
		"""List of lengths of the input variables, should match with the input variables"""
		return interface.GetArrayValue(self._id, 'correlation_lengths')

	@correlation_lengths.setter
	def correlation_lengths(self, values : list[float]):
		interface.SetArrayValue(self._id, 'correlation_lengths', values)

	@property
	def length(self) -> float:
		"""Length to be applied to the design point. The resulting design point will be applicable to this length"""
		return interface.GetValue(self._id, 'length')

	@length.setter
	def length(self, value: float):
		interface.SetValue(self._id, 'length', value)

	def run(self):
		"""Applies the length effect to the `design_point_cross_section`.
        Results is a design point, which is part of this project. When failed, the `design_point` is empty."""
		self._design_point = None
		interface.SetIntValue(self._id, 'design_point_cross_section', self._design_point_cross_section._id)
		interface.SetIntValue(self._id, 'correlation_matrix', self._correlation_matrix._id)
		interface.Execute(self._id, 'run')

	@property
	def design_point(self) -> DesignPoint:
		"""The length effect applied design point, invoked by `run`
        Is empty when the run failed. The original `design_point_cross_section` is added to the `DesignPoint.contributing_design_point`"""
		if self._design_point is None:
			designPointId = interface.GetIdValue(self._id, 'design_point')
			if designPointId > 0:
				variables = self._design_point_cross_section.get_variables()
				self._design_point = DesignPoint(designPointId, variables, self._design_point_cross_section)
		return self._design_point




# Copyright (C) Stichting Deltares. All rights reserved.
#
# This file is part of the Probabilistic Library.
#
# The Probabilistic Library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# All names, logos, and references to "Deltares" are registered trademarks of
# Stichting Deltares and remain full property of Stichting Deltares at all times.
# All rights reserved.
#


import sys
from math import isnan
from enum import Enum
import matplotlib.pyplot as plt

from .utils import FrozenObject, FrozenList, CallbackList
from .logging import Evaluation, Message, ValidationReport
from .statistic import Stochast, ProbabilityValue
from .reliability import StochastSettings, GradientType
from . import interface

if not interface.IsLibraryLoaded():
	interface.LoadDefaultLibrary()

class UncertaintyMethod(Enum):
	"""Enumeration which defines the algorithm to perform an uncertainty analysis"""
	form = 'form'
	fosm = 'fosm'
	numerical_integration = 'numerical_integration'
	crude_monte_carlo = 'crude_monte_carlo'
	importance_sampling = 'importance_sampling'
	directional_sampling = 'directional_sampling'
	def __str__(self):
		return str(self.value)

class UncertaintySettings(FrozenObject):
	"""Settings of an uncertainty algorithm

    Settings of all uncertainty algorithms are combined in this class. Often settings only apply to
    a selected number of algorithms. Settings per variable are listed in `stochast_settings`.

    The settings are divided into the following categories:

    Algorithm settings: Settings which are used to control the algorithm, such as the number of allowed
    samples and the convergence criterion.

    Runtime settings: Settings to control how model executions take place and which additional output to produce.
    These settings do not influence the calculated uncertainty."""

	def __init__(self):
		self._id = interface.Create('uncertainty_settings')
		self._stochast_settings = FrozenList()
		self._quantiles = None
		self._synchronizing = False
		super()._freeze()
		interface.SetBoolValue(self._id, 'use_openmp_in_reliability', False)

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['max_parallel_processes',
		        'save_realizations',
		        'save_messages',
		        'reuse_calculations'
		        'uncertainty_method',
				'is_repeatable_random',
				'random_seed',
	            'minimum_samples',
	            'maximum_samples',
	            'maximum_iterations',
	            'minimum_directions',
	            'maximum_directions',
	            'minimum_u',
	            'maximum_u',
	            'step_size',
		        'gradient_type',
	            'global_step_size',
	            'variation_coefficient',
			    'probability_for_convergence',
			    'derive_samples_from_variation_coefficient',
			    'calculate_correlations',
			    'calculate_input_correlations',
			    'quantiles',
	            'stochast_settings',
		        'is_valid',
		        'validate']


	@property
	def max_parallel_processes(self) -> int:
		"""The number of parallel executions of model evaluations"""
		return interface.GetIntValue(self._id, 'max_parallel_processes')

	@max_parallel_processes.setter
	def max_parallel_processes(self, value : int):
		interface.SetIntValue(self._id, 'max_parallel_processes', value)

	@property
	def save_realizations(self) -> bool:
		"""Indicates whether samples should be saved
        If saved, the samples will be part of the design point"""
		return interface.GetBoolValue(self._id, 'save_realizations')

	@save_realizations.setter
	def save_realizations(self, value : bool):
		interface.SetBoolValue(self._id, 'save_realizations', value)

	@property
	def save_messages(self) -> bool:
		"""Indicates whether messages generated by the uncertainty analysis will be saved
        If saved, messages will be part of the design point"""
		return interface.GetBoolValue(self._id, 'save_messages')

	@save_messages.setter
	def save_messages(self, value : bool):
		interface.SetBoolValue(self._id, 'save_messages', value)

	@property
	def reuse_calculations(self) -> bool:
		"""Indicates whether prior model results will be reused by the uncertainty analysis.

        This will speed up calculations when several analyses are performed, for which the same realizations
        will have to be executed, for example a Crude Monte Carlo analysis for different output parameters.
        But when a modification to the model is made, which is beyond the scope of the  model definition,
        this leads to undesired results"""

		return interface.GetBoolValue(self._id, 'reuse_calculations')

	@reuse_calculations.setter
	def reuse_calculations(self, value : bool):
		interface.SetBoolValue(self._id, 'reuse_calculations', value)

	@property
	def uncertainty_method(self) -> UncertaintyMethod:
		"""Defines the uncertainty algorithm"""
		return UncertaintyMethod[interface.GetStringValue(self._id, 'uncertainty_method')]

	@uncertainty_method.setter
	def uncertainty_method(self, value : UncertaintyMethod):
		interface.SetStringValue(self._id, 'uncertainty_method', str(value))

	@property
	def is_repeatable_random(self) -> bool:
		"""Indicates whether in each run the same random samples will be generated"""
		return interface.GetBoolValue(self._id, 'is_repeatable_random')

	@is_repeatable_random.setter
	def is_repeatable_random(self, value : bool):
		interface.SetBoolValue(self._id, 'is_repeatable_random', value)

	@property
	def random_seed(self) -> int:
		"""Seed number for the random generator"""
		return interface.GetIntValue(self._id, 'random_seed')

	@random_seed.setter
	def random_seed(self, value : int):
		interface.SetIntValue(self._id, 'random_seed', value)

	@property
	def minimum_samples(self) -> int:
		"""The minimum number of samples to be used"""
		return interface.GetIntValue(self._id, 'minimum_samples')

	@minimum_samples.setter
	def minimum_samples(self, value : int):
		interface.SetIntValue(self._id, 'minimum_samples', value)

	@property
	def maximum_samples(self) -> int:
		"""The maximum number of samples to be used"""
		return interface.GetIntValue(self._id, 'maximum_samples')

	@maximum_samples.setter
	def maximum_samples(self, value : int):
		interface.SetIntValue(self._id, 'maximum_samples', value)

	@property
	def maximum_iterations(self) -> int:
		"""The maximum number of iterations to be used, only for FORM"""
		return interface.GetIntValue(self._id, 'maximum_iterations')

	@maximum_iterations.setter
	def maximum_iterations(self, value : int):
		interface.SetIntValue(self._id, 'maximum_iterations', value)

	@property
	def minimum_directions(self) -> int:
		"""The minimum number of directions to be used, only for directional sampling"""
		return interface.GetIntValue(self._id, 'minimum_directions')

	@minimum_directions.setter
	def minimum_directions(self, value : int):
		interface.SetIntValue(self._id, 'minimum_directions', value)

	@property
	def maximum_directions(self) -> int:
		"""The maximum number of directions to be used, only for directional sampling"""
		return interface.GetIntValue(self._id, 'maximum_directions')

	@maximum_directions.setter
	def maximum_directions(self, value : int):
		interface.SetIntValue(self._id, 'maximum_directions', value)

	@property
	def minimum_u(self) -> float:
		"""The minimum u-value to be applied to a stochastic variable, used for FORM"""
		return interface.GetValue(self._id, 'minimum_u')

	@minimum_u.setter
	def minimum_u(self, value : float):
		interface.SetValue(self._id, 'minimum_u', value)

	@property
	def maximum_u(self) -> float:
		"""The maximum u-value to be applied to a stochastic variable, used for FORM"""
		return interface.GetValue(self._id, 'maximum_u')

	@maximum_u.setter
	def maximum_u(self, value : float):
		interface.SetValue(self._id, 'maximum_u', value)

	@property
	def global_step_size(self) -> float:
		"""Step size between fragility values, which is the result if a FORM calculation"""
		return interface.GetValue(self._id, 'global_step_size')

	@global_step_size.setter
	def global_step_size(self, value : float):
		interface.SetValue(self._id, 'global_step_size', value)

	@property
	def step_size(self) -> float:
		"""Step size in calculating the model gradient, used by FORM"""
		return interface.GetValue(self._id, 'step_size')

	@step_size.setter
	def step_size(self, value : float):
		interface.SetValue(self._id, 'step_size', value)

	@property
	def gradient_type(self) -> GradientType:
		"""Method to determine the gradient of a model"""
		return GradientType[interface.GetStringValue(self._id, 'gradient_type')]

	@gradient_type.setter
	def gradient_type(self, value : GradientType):
		interface.SetStringValue(self._id, 'gradient_type', str(value))

	@property
	def variance_factor(self) -> float:
		"""Variance factor, used by importance sampling"""
		return interface.GetValue(self._id, 'variance_factor')

	@variance_factor.setter
	def variance_factor(self, value : float):
		interface.SetValue(self._id, 'variance_factor', value)

	@property
	def variation_coefficient(self) -> float:
		"""Convergence criterion, used by Monte Carlo family algorithms"""
		return interface.GetValue(self._id, 'variation_coefficient')

	@variation_coefficient.setter
	def variation_coefficient(self, value : float):
		interface.SetValue(self._id, 'variation_coefficient', value)

	@property
	def probability_for_convergence(self) -> float:
		"""Probability to which the convergence applies

        The convergence is calculated as if it were a reliability analysis. The reliability analysis is tuned
        in such a way (by modification of the failure definition) that it produces the given probability."""

		return interface.GetValue(self._id, 'probability_for_convergence')

	@probability_for_convergence.setter
	def probability_for_convergence(self, value : float):
		interface.SetValue(self._id, 'probability_for_convergence', value)

	@property
	def derive_samples_from_variation_coefficient(self) -> bool:
		"""Indicates that the number of samples is derived from the variation coefficient and the probability of
        failure (only for Crude Monte Carlo)"""
		return interface.GetBoolValue(self._id, 'derive_samples_from_variation_coefficient')

	@derive_samples_from_variation_coefficient.setter
	def derive_samples_from_variation_coefficient(self, value : bool):
		interface.SetBoolValue(self._id, 'derive_samples_from_variation_coefficient', value)

	@property
	def calculate_correlations(self) -> bool:
		"""Indicates that correlations between output parameters must be calculated"""
		return interface.GetBoolValue(self._id, 'calculate_correlations')

	@calculate_correlations.setter
	def calculate_correlations(self, value : bool):
		interface.SetBoolValue(self._id, 'calculate_correlations', value)

	@property
	def calculate_input_correlations(self) -> bool:
		"""Indicates that correlations between output parameters and input parameters must be calculated"""
		return interface.GetBoolValue(self._id, 'calculate_input_correlations')

	@calculate_input_correlations.setter
	def calculate_input_correlations(self, value : bool):
		interface.SetBoolValue(self._id, 'calculate_input_correlations', value)

	@property
	def stochast_settings(self) -> list[StochastSettings]:
		"""List of settings specified per stochastic variable"""
		return self._stochast_settings

	@property
	def quantiles(self) -> list[ProbabilityValue]:
		"""List of quantiles of the output parameter, which must be calculated"""
		if self._quantiles is None:
			self._synchronizing = True
			self._quantiles = CallbackList(self._quantiles_changed)
			quantile_ids = interface.GetArrayIntValue(self._id, 'quantiles')
			for quantile_id in quantile_ids:
				self._quantiles.append(ProbabilityValue(quantile_id))
			self._synchronizing = False

		return self._quantiles

	def _quantiles_changed(self):
		if not self._synchronizing:
			# replace floats by ProbabilityValue
			self._synchronizing = True
			for i in range(len(self._quantiles)):
				if type(self._quantiles[i]) == int or type(self._quantiles[i]) == float:
					val = self._quantiles[i]
					self._quantiles[i] = ProbabilityValue()
					self._quantiles[i].probability_of_non_failure = val
			self._synchronizing = False

			interface.SetArrayIntValue(self._id, 'quantiles', [quantile._id for quantile in self._quantiles])

	def is_valid(self) -> bool:
		"""Indicates whether the settings are valid"""
		return interface.GetBoolValue(self._id, 'is_valid')

	def validate(self):
		"""Prints the validity of the settings"""
		id_ = interface.GetIdValue(self._id, 'validate')
		if id_ > 0:
			validation_report = ValidationReport(id_)
			validation_report.print()

	def _set_variables(self, variables):
		new_stochast_settings = []
		for variable in variables:
			stochast_setting = self._stochast_settings[str(variable)]
			if stochast_setting is None:
				stochast_setting = StochastSettings(variable)
			new_stochast_settings.append(stochast_setting)
		self._stochast_settings = FrozenList(new_stochast_settings)
		interface.SetArrayIntValue(self._id, 'stochast_settings', [stochast_setting._id for stochast_setting in self._stochast_settings])

class UncertaintyResult(FrozenObject):
	"""Contains the result of an uncertainty analysis for a specific output parameter

    An uncertainty analysis results in a stochastic variable. The characteristics of this stochastic variable describe the
    uncertainty of the output parameter. Depending on the uncertainty algorithm different distributions may be used."""

	def __init__(self, id : int, variables : list[Stochast]):
		self._id = id
		self._variable = None
		self._messages = None
		self._realizations = None
		self._quantile_realizations = None
		self._variables = FrozenList(variables)
		super()._freeze()

	def __del__(self):
		try:
			interface.Destroy(self._id)
		except:
			pass

	def __dir__(self):
		return ['identifier',
		        'variable',
		        'quantile_realizations',
		        'realizations',
		        'messages',
		        'print',
		        'plot',
		        'plot_realizations',
		        'get_plot',
		        'get_plot_realizations']

	def __str__(self):
		return self.identifier

	@property
	def identifier(self) -> str:
		"""Identification of the output parameter"""
		return interface.GetStringValue(self._id, 'identifier')

	@property
	def variable(self) -> Stochast:
		"""Stochastic variable describing the uncertainty of the output parameter"""
		if self._variable is None:
			variable_id = interface.GetIdValue(self._id, 'variable')
			if variable_id > 0:
				self._variable = Stochast(variable_id);

		return self._variable

	@property
	def realizations(self) -> list[Evaluation]:
		"""List of samples calculated by the uncertainty algorithm. Depends on the setting `UncertaintySettings.save_realizations` whether
        this list is provided"""
		if self._realizations is None:
			realizations = []
			realization_ids = interface.GetArrayIdValue(self._id, 'evaluations')
			for realization_id in realization_ids:
				realizations.append(Evaluation(realization_id))
			self._realizations = FrozenList(realizations)

		return self._realizations

	@property
	def quantile_realizations(self) -> list[Evaluation]:
		"""List of samples corresponding with the list of quantiles in the uncertainty settings.
       The samples in this list are the closest samples to the given quantiles"""
		if self._quantile_realizations is None:
			quantile_realizations = []
			quantile_realization_ids = interface.GetArrayIdValue(self._id, 'quantile_evaluations')
			for realization_id in quantile_realization_ids:
				quantile_realizations.append(Evaluation(realization_id))
			self._quantile_realizations = FrozenList(quantile_realizations)

		return self._quantile_realizations

	@property
	def messages(self) -> list[Message]:
		"""List of messages generated by the reliability algorithm. Depends on the setting `UncertaintySettings.save_messages` whether this
        list is provided"""
		if self._messages is None:
			messages = []
			message_ids = interface.GetArrayIdValue(self._id, 'messages')
			for message_id in message_ids:
				messages.append(Message(message_id))
			self._messages = FrozenList(messages)

		return self._messages

	def print(self, decimals=4):
		"""Prints the resulting stochastic variable and quantile samples.

        Parameters
        ----------
        decimals : int, optional
            The number of decimals to print"""
		self.variable.print(decimals)
		if len(self.quantile_realizations) > 0:
			print('Quantiles:')
			for quantile in self.quantile_realizations:
				quantile._print(1, decimals)

	def plot(self, xmin : float = None, xmax : float = None):
		"""Shows a plot of the resulting stochastic variable"""
		self.get_plot(xmin, xmax).show()

	def get_plot(self, xmin : float = None, xmax : float = None) -> plt:
		"""Gets a plot object of the resulting stochastic variable"""

		vplot = self.variable.get_plot(xmin, xmax)

		plot_legend = False
		for ii in range(len(self.quantile_realizations)):
			vplot.axvline(x=self.quantile_realizations[ii].output_values[0], color="green", linestyle="--", label=f"{self.quantile_realizations[ii].quantile:.4g}-quantile")
			plot_legend = True

		if plot_legend:
			vplot.legend()

		return vplot

	def plot_realizations(self, var_x : str | Stochast = None, var_y : str | Stochast = None):
		"""Shows a scatter-plot of realizations performed by the uncertainty analysis. The x-y coordinates
        correspond with realization input values. Only available when `UncertaintySettings.save_realizations`
        was set in the `UncertaintySettings`.

        Parameters
        ----------
        var_x : str | Stochast, optional
            The stochastic variable to use for the x-axis, if omitted the first variable is used

        var_y : str | Stochast, optional
            The stochastic variable to use for the y-axis, if omitted the second variable is used"""

		self.get_plot_realizations(var_x, var_y).show()

	def get_plot_realizations(self, var_x : str | Stochast = None, var_y : str | Stochast = None) -> plt:
		"""Gets a plot object of a scatter-plot of realizations performed by the uncertainty analysis. The
        x-y coordinates correspond with realization input values. Only available when `UncertaintySettings.save_realizations`
        was set in the `UncertaintySettings`.

        Parameters
        ----------
        var_x : str | Stochast, optional
            The stochastic variable to use for the x-axis, if omitted the first variable is used.

        var_y : str | Stochast, optional
            The stochastic variable to use for the y-axis, if omitted the second variable is used"""

		if len(self.realizations) == 0:
			print ("No realizations were saved, run again with settings.save_realizations = True")
			return plt

		if len(self._variables) < 2:
			print ("Not enough variables to plot realizations")
			return plt

		plt.close()

		# the first two variables
		index = [0, 1]

		if var_x is not None:
			index[0] = self._variables.index(var_x)
		if var_y is not None:
			index[1] = self._variables.index(var_y)

		if index[0] < 0 or index[1] < 0:
			print ("Variables could not be found")
			return

		x_values = [realization.input_values[index[0]] for realization in self.realizations]
		y_values = [realization.input_values[index[1]] for realization in self.realizations]

		# plot realizations
		plt.figure()
		plt.grid(True)
		plt.scatter(x_values, y_values, color='b')

		x_values = [realization.input_values[index[0]] for realization in self.quantile_realizations]
		y_values = [realization.input_values[index[1]] for realization in self.quantile_realizations]

		plt.scatter(x_values, y_values, color='black')

		plt.xlabel(self._variables[index[0]].name)
		plt.ylabel(self._variables[index[1]].name)

		plt.title('Realizations', fontsize=14)

		return plt



# Copyright (C) Stichting Deltares. All rights reserved.
#
# This file is part of the Probabilistic Library.
#
# The Probabilistic Library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# All names, logos, and references to "Deltares" are registered trademarks of
# Stichting Deltares and remain full property of Stichting Deltares at all times.
# All rights reserved.
#
import sys
from typing import Iterator, TypeVar, Generic

T = TypeVar('T')

class CallbackList(list):
	"""List which invokes a callback after each list modification"""

	def __init__(self, callback):
		"""Constructor

        Parameters
        ----------
        callback : method
            Method to invoke after each list modification"""

		self._callback = callback

	def __setitem__(self, index, item):
		super().__setitem__(index, item)
		self._callback()
		
	def clear(self):
		super().clear()
		self._callback()
		
	def append(self, item):
		super().append(item)
		self._callback()
		
	def insert(self, index, item):
		super().insert(index, item)
		self._callback()
		
	def remove(self, item):
		super().remove(item)
		self._callback()

	def pop(self, item):
		item = super().pop(item)
		self._callback()
		return item

	def extend(self, items):
		super().extend(items)
		self._callback()

class FrozenList(Generic[T]):
	"""Read-only list, items can also be retrieved by their string representation"""

	def __init__(self, initial_list = None):
		"""Constructor

        Parameters
        ----------
        initial_list : list, optional
            List which will be wrapped as a read-only list by this class"""

		self._list : list[T] = []
		self._dict : dict[str, T] = {}
		if not initial_list is None:
			self._list.extend(initial_list)
			for item in self._list:
				self._dict[str(item)] = item

	def __getitem__(self, index) -> T:
		if isinstance(index, int):
			return self._list[index]
		elif isinstance(index, slice):
			return FrozenList(self._list[index])
		else:
			if not isinstance(index, str):
				index = str(index)
			if index in self._dict.keys():
				return self._dict[index]
			else:
				return None

	def __iter__(self) -> Iterator[T]:
		return self._list.__iter__()

	def __next__(self) -> T:
		self._list.__next__()

	def __len__(self) -> int:
		return len(self._list)

	def __str__(self) -> str:
		return str(self._list)

	def index(self, item, start = 0, stop = sys.maxsize) -> int:
		"""Gets the index of an item

        Parameters
        ----------
        item : obj
            Item to be found
        start : int, optional
            Start index
        stop : int, optional
            Stop index"""

		if isinstance(item, str):
			item = self[item]
		if item != None:
			return self._list.index(item, start, stop)
		else:
			return -1

	def count(self) -> int:
		"""Gets the number of items in the list"""
		return self._list.count()

	def get_list(self) -> list[T]:
		"""Gets a copy of the list with full access"""
		getlist = []
		getlist.extend(self._list)
		return getlist

class FrozenObject:
	"""Object to which no members can be added. If members are added, an exception occurs."""
	def __setattr__(self, key, value):
		if hasattr(self, '_frozen'):
			if key in self.__dir__() or hasattr(self, key):
				super.__setattr__(self, key, value)
			else:
				raise ValueError(key + ' does not exist')
		else:
			super.__setattr__(self, key, value)

	def __dir__(self):
		return []

	def _freeze(self):
		self._frozen = True

class PrintUtils:
	"""Utilities for printing"""
	def get_space_from_indent(indent : int) -> str:
		"""Converts indentation level to spaces"""
		indent_str = ''
		for i in range(indent):
			indent_str += '  '
		return indent_str

class NumericUtils:
	"""Numeric utilities"""
	def order (value1 : float, value2 : float) -> tuple[float, float]:
		"""Returns floats ordered"""
		if value1 > value2:
			return value2, value1
		else:
			return value1, value2

	def make_different(value1 : float, value2 : float) -> tuple[float, float]:
		"""Guarantees that values are different"""
		if value1 == value2:
			diff = abs(value1) / 10
			if diff == 0:
				diff = 1
			value1 = value1 - diff
			value2 = value2 + diff
		return value1, value2



# Copyright (C) Stichting Deltares. All rights reserved.
#
# This file is part of the Probabilistic Library.
#
# The Probabilistic Library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# All names, logos, and references to "Deltares" are registered trademarks of
# Stichting Deltares and remain full property of Stichting Deltares at all times.
# All rights reserved.
#

from __future__ import annotations
from math import isnan
import matplotlib.pyplot as plt
import sys
from enum import Enum

from .utils import FrozenObject, FrozenList, PrintUtils, CallbackList
from .statistic import Stochast, FragilityValue
from .logging import Message, Evaluation, ValidationReport
from . import interface

if not interface.IsLibraryLoaded():
	interface.LoadDefaultLibrary()

class ReliabilityMethod(Enum):
	"""Enumeration which defines the algorithm to perform a reliability analysis"""
	form = 'form'
	numerical_integration = 'numerical_integration'
	crude_monte_carlo = 'crude_monte_carlo'
	importance_sampling = 'importance_sampling'
	adaptive_importance_sampling = 'adaptive_importance_sampling'
	directional_sampling = 'directional_sampling'
	subset_simulation = 'subset_simulation'
	numerical_bisection = 'numerical_bisection'
	latin_hypercube = 'latin_hypercube'
	cobyla_reliability = 'cobyla_reliability'
	form_then_directional_sampling = 'form_then_directional_sampling'
	directional_sampling_then_form = 'directional_sampling_then_form'
	def __str__(self):
		return str(self.value)

class DesignPointMethod(Enum):
	"""Enumeration which defines how the design point is determined"""
	nearest_to_mean = 'nearest_to_mean'
	center_of_gravity = 'center_of_gravity'
	center_of_angles = 'center_of_angles'
	def __str__(self):
		return str(self.value)

class StartMethod(Enum):
	"""Enumeration which defines the starting point of some reliability algorithms"""
	fixed_value = 'fixed_value'
	one = 'one'
	ray_search = 'ray_search'
	sphere_search = 'sphere_search'
	sensitivity_search = 'sensitivity_search'
	def __str__(self):
		return str(self.value)

class GradientType(Enum):
	"""Enumeration which defines how a gradient is calculated"""
	single = 'single'
	double = 'double'
	def __str__(self):
		return str(self.value)

class SampleMethod(Enum):
	"""Enumeration which defines how samples are generated in the subset simulation algorithm"""
	markov_chain = 'markov_chain'
	adaptive_conditional = 'adaptive_conditional'
	def __str__(self):
		return str(self.value)

class CombinerMethod(Enum):
	"""Enumeration which defines the algorithm to combine design points"""
	hohenbichler = 'hohenbichler'
	importance_sampling = 'importance_sampling'
	directional_sampling = 'directional_sampling'
	hohenbichler_form = 'hohenbichler_form'
	def __str__(self):
		return str(self.value)

class ExcludingCombinerMethod(Enum):
	"""Enumeration which defines the algorithm to combine excluding design points"""
	weighted_sum = 'weighted_sum'
	hohenbichler = 'hohenbichler_excluding'
	def __str__(self):
		return str(self.value)

class CombineType(Enum):
	"""Enumeration which defines how design points are combined"""

	series = 'series'
	"""OR port, the result of the combination is the probability that one of the failure conditions will happen"""

	parallel = 'parallel'
	"""AND port, the result of the combination is the probability that all of the failure conditions will happen."""
	def __str__(self):
		return str(self.value)

class MessageType(Enum):
	"""Indicates the severity of a message
    Only in case of message type 'error' a calculation is not allowed or has failed"""

	debug = 'debug'
	info = 'info'
	warning = 'warning'
	error = 'error'
	def __str__(self):
		return str(self.value)

class Settings(FrozenObject):
	"""Settings of a reliability algorithm

    Settings of all reliability algorithms are combined in this class. Often settings only apply to
    a selected number of algorithms. Settings per variable are listed in `stochast_settings`.

    The settings are divided into the following categories:

    Algorithm settings: Settings which are used to control the algorithm, such as the number of allowed
    samples and the convergence criterion.

    Runtime settings: Settings to control how model executions take place and which additional output to produce.
    These settings do not influence the calculated reliability index and alpha values."""

	def __init__(self, id_ = None):
		if id_ is None:
			self._id = interface.Create('settings')
		else:
			self._id = id_
		self._stochast_settings = FrozenList()
		super()._freeze()
		interface.SetBoolValue(self._id, 'use_openmp_in_reliability', False)

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['max_parallel_processes',
				'save_realizations',
				'save_convergence',
				'save_messages',
		        'reuse_calculations',
				'reliability_method',
				'design_point_method',
				'start_method',
				'all_quadrants',
				'max_steps_sphere_search',
				'is_repeatable_random',
				'random_seed',
				'sample_method'
				'minimum_samples',
				'maximum_samples',
				'minimum_iterations',
				'maximum_iterations',
				'minimum_directions',
				'maximum_directions',
				'epsilon_beta',
				'step_size',
				'gradient_type',
				'relaxation_factor',
				'relaxation_loops',
				'minimum_variance_loops',
				'maximum_variance_loops',
				'variation_coefficient',
				'fraction_failed',
				'stochast_settings',
				'is_valid',
				'validate']

	@property
	def max_parallel_processes(self) -> int:
		"""The number of parallel executions of model evaluations"""
		return interface.GetIntValue(self._id, 'max_parallel_processes')

	@max_parallel_processes.setter
	def max_parallel_processes(self, value : int):
		interface.SetIntValue(self._id, 'max_parallel_processes', value)

	@property
	def save_realizations(self) -> bool:
		"""Indicates whether samples should be saved
        If saved, the samples will be part of the design point"""
		return interface.GetBoolValue(self._id, 'save_realizations')

	@save_realizations.setter
	def save_realizations(self, value : bool):
		interface.SetBoolValue(self._id, 'save_realizations', value)

	@property
	def save_convergence(self) -> bool:
		"""Indicates whether information about convergence of the reliability analysis should be saved
        If saved, convergence data will be part of the design point"""
		return interface.GetBoolValue(self._id, 'save_convergence')

	@save_convergence.setter
	def save_convergence(self, value : bool):
		interface.SetBoolValue(self._id, 'save_convergence', value)

	@property
	def save_messages(self) -> bool:
		"""Indicates whether messages generated by the reliability analysis will be saved
        If saved, messages will be part of the design point"""
		return interface.GetBoolValue(self._id, 'save_messages')

	@save_messages.setter
	def save_messages(self, value : bool):
		interface.SetBoolValue(self._id, 'save_messages', value)

	@property
	def reuse_calculations(self) -> bool:
		"""Indicates whether prior model results will be reused by the reliability analysis.

        This will speed up calculations when several analyses are performed, for which the same realizations
        will have to be executed, for example a Crude Monte Carlo analysis with different limit state
        definitions. But when a modification to the model is made, which is beyond the scope of the
        model definition, this leads to undesired results"""

		return interface.GetBoolValue(self._id, 'reuse_calculations')

	@reuse_calculations.setter
	def reuse_calculations(self, value : bool):
		interface.SetBoolValue(self._id, 'reuse_calculations', value)

	@property
	def reliability_method(self) -> ReliabilityMethod:
		"""Defines the reliability algorithm"""
		return ReliabilityMethod[interface.GetStringValue(self._id, 'reliability_method')]

	@reliability_method.setter
	def reliability_method(self, value : ReliabilityMethod):
		interface.SetStringValue(self._id, 'reliability_method', str(value))

	@property
	def design_point_method(self) -> DesignPointMethod:
		"""Defines the algorithm how the design point is derived"""
		return DesignPointMethod[interface.GetStringValue(self._id, 'design_point_method')]

	@design_point_method.setter
	def design_point_method(self, value : DesignPointMethod):
		interface.SetStringValue(self._id, 'design_point_method', str(value))

	@property
	def start_method(self) -> StartMethod:
		"""Defines the starting point of the algorithm"""
		return StartMethod[interface.GetStringValue(self._id, 'start_method')]

	@start_method.setter
	def start_method(self, value : StartMethod):
		interface.SetStringValue(self._id, 'start_method', str(value))

	@property
	def all_quadrants(self) -> bool:
		"""Indicates whether all quadrants should be evaluated, only used by sphere search"""
		return interface.GetBoolValue(self._id, 'all_quadrants')

	@all_quadrants.setter
	def all_quadrants(self, value : bool):
		interface.SetBoolValue(self._id, 'all_quadrants', value)

	@property
	def max_steps_sphere_search(self) -> int:
		"""The maximum number of steps in sphere search to determine the starting point"""
		return interface.GetIntValue(self._id, 'max_steps_sphere_search')

	@max_steps_sphere_search.setter
	def max_steps_sphere_search(self, value : int):
		interface.SetIntValue(self._id, 'max_steps_sphere_search', value)

	@property
	def is_repeatable_random(self) -> bool:
		"""Indicates whether in each run the same random samples will be generated"""
		return interface.GetBoolValue(self._id, 'is_repeatable_random')

	@is_repeatable_random.setter
	def is_repeatable_random(self, value : bool):
		interface.SetBoolValue(self._id, 'is_repeatable_random', value)

	@property
	def random_seed(self) -> int:
		"""Seed number for the random generator"""
		return interface.GetIntValue(self._id, 'random_seed')

	@random_seed.setter
	def random_seed(self, value : int):
		interface.SetIntValue(self._id, 'random_seed', value)

	@property
	def sample_method(self) -> SampleMethod:
		"""The way samples are generated for the subset simulation algorithm"""
		return SampleMethod[interface.GetStringValue(self._id, 'sample_method')]

	@sample_method.setter
	def sample_method(self, value : SampleMethod):
		interface.SetStringValue(self._id, 'sample_method', str(value))

	@property
	def minimum_samples(self) -> int:
		"""The minimum number of samples to be used"""
		return interface.GetIntValue(self._id, 'minimum_samples')

	@minimum_samples.setter
	def minimum_samples(self, value : int):
		interface.SetIntValue(self._id, 'minimum_samples', value)

	@property
	def maximum_samples(self) -> int:
		"""The maximum number of samples to be used"""
		return interface.GetIntValue(self._id, 'maximum_samples')

	@maximum_samples.setter
	def maximum_samples(self, value : int):
		interface.SetIntValue(self._id, 'maximum_samples', value)

	@property
	def minimum_iterations(self) -> int:
		"""The minimum number of iterations to be used"""
		return interface.GetIntValue(self._id, 'minimum_iterations')

	@minimum_iterations.setter
	def minimum_iterations(self, value : int):
		interface.SetIntValue(self._id, 'minimum_iterations', value)

	@property
	def maximum_iterations(self) -> int:
		"""The maximum number of iterations to be used"""
		return interface.GetIntValue(self._id, 'maximum_iterations')

	@maximum_iterations.setter
	def maximum_iterations(self, value : int):
		interface.SetIntValue(self._id, 'maximum_iterations', value)

	@property
	def minimum_directions(self) -> int:
		"""The minimum number of directions to be used, only for directional sampling"""
		return interface.GetIntValue(self._id, 'minimum_directions')

	@minimum_directions.setter
	def minimum_directions(self, value : int):
		interface.SetIntValue(self._id, 'minimum_directions', value)

	@property
	def maximum_directions(self) -> int:
		"""The maximum number of directions to be used, only for directional sampling"""
		return interface.GetIntValue(self._id, 'maximum_directions')

	@maximum_directions.setter
	def maximum_directions(self, value : int):
		interface.SetIntValue(self._id, 'maximum_directions', value)

	@property
	def epsilon_beta(self) -> float:
		"""Convergence criterion for FORM, using the maximum allowed predicted uncertainty in reliability index"""
		return interface.GetValue(self._id, 'epsilon_beta')

	@epsilon_beta.setter
	def epsilon_beta(self, value : float):
		interface.SetValue(self._id, 'epsilon_beta', value)

	@property
	def step_size(self) -> float:
		"""Step size in finding the gradient of a model, defined in u-space"""
		return interface.GetValue(self._id, 'step_size')

	@step_size.setter
	def step_size(self, value : float):
		interface.SetValue(self._id, 'step_size', value)

	@property
	def gradient_type(self) -> GradientType:
		"""Method to determine the gradient of a model"""
		return GradientType[interface.GetStringValue(self._id, 'gradient_type')]

	@gradient_type.setter
	def gradient_type(self, value : GradientType):
		interface.SetStringValue(self._id, 'gradient_type', str(value))

	@property
	def relaxation_factor(self) -> float:
		"""Relaxation factor, used by FORM"""
		return interface.GetValue(self._id, 'relaxation_factor')

	@relaxation_factor.setter
	def relaxation_factor(self, value : float):
		interface.SetValue(self._id, 'relaxation_factor', value)

	@property
	def relaxation_loops(self) -> int:
		"""Number of relaxation loops. In case of no convergence, the relaxation factor is decreased until convergence is found, used by FORM"""
		return interface.GetIntValue(self._id, 'relaxation_loops')

	@relaxation_loops.setter
	def relaxation_loops(self, value):
		interface.SetIntValue(self._id, 'relaxation_loops', value)

	@property
	def maximum_variance_loops(self) -> int:
		"""Maximum number of importance sampling loops, used by adaptive importance sampling"""
		return interface.GetIntValue(self._id, 'maximum_variance_loops')

	@maximum_variance_loops.setter
	def maximum_variance_loops(self, value : int):
		interface.SetIntValue(self._id, 'maximum_variance_loops', value)

	@property
	def minimum_variance_loops(self) -> int:
		"""Minimum number of importance sampling loops, used by adaptive importance sampling"""
		return interface.GetIntValue(self._id, 'minimum_variance_loops')

	@minimum_variance_loops.setter
	def minimum_variance_loops(self, value : int):
		interface.SetIntValue(self._id, 'minimum_variance_loops', value)

	@property
	def variation_coefficient(self) -> float:
		"""Convergence criterion, used by Monte Carlo family algorithms"""
		return interface.GetValue(self._id, 'variation_coefficient')

	@variation_coefficient.setter
	def variation_coefficient(self, value : float):
		interface.SetValue(self._id, 'variation_coefficient', value)

	@property
	def fraction_failed(self) -> float:
		"""Indicates the fraction of failed samples. Criterion for adaptive importance sampling whether to perform
        another importance sampling loop"""
		return interface.GetValue(self._id, 'fraction_failed')

	@fraction_failed.setter
	def fraction_failed(self, value : float):
		interface.SetValue(self._id, 'fraction_failed', value)

	@property
	def stochast_settings(self) -> list[StochastSettings]:
		"""List of settings specified per stochastic variable"""
		return self._stochast_settings

	def _set_variables(self, variables):
		new_stochast_settings = []
		for variable in variables:
			stochast_setting = self._stochast_settings[str(variable)]
			if stochast_setting is None:
				stochast_setting = StochastSettings(variable)
			new_stochast_settings.append(stochast_setting)
		self._stochast_settings = FrozenList(new_stochast_settings)
		interface.SetArrayIntValue(self._id, 'stochast_settings', [stochast_setting._id for stochast_setting in self._stochast_settings])

	def is_valid(self) -> bool:
		"""Indicates whether the settings are valid"""
		return interface.GetBoolValue(self._id, 'is_valid')

	def validate(self):
		"""Prints the validity of the settings"""
		id_ = interface.GetIdValue(self._id, 'validate')
		if id_ > 0:
			validation_report = ValidationReport(id_)
			validation_report.print()

class StochastSettings(FrozenObject):
	"""Defines reliability or uncertainty settings for a stochastic variable

    These settings are part of reliability settings or uncertainty settings. When using a project, these
    settings are generated automatically."""

	def __init__(self, variable):
		self._id = interface.Create('stochast_settings')
		self._variable = variable
		if not variable is None:
			interface.SetIntValue(self._id, 'variable', self._variable._id)
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['min_value',
				'max_value',
				'start_value',
				'variance_factor',
				'is_initialization_allowed',
				'is_variance_allowed',
				'intervals']

	def __str__(self):
		if self._variable is None:
			return ''
		else:
			return self._variable.name

	@property
	def variable(self) -> Stochast:
		"""The stochastic variable to which these settings apply
        This value cannot be set. A project generates these settings automatically when a model is set to the project."""

		if self._variable is None:
			id_ = interface.GetIdValue(self._id, 'variable')
			if id_ > 0:
				self._variable = Stochast(id_)
		return self._variable

	@property
	def min_value(self) -> float:
		"""The minimum value which will be assigned to the stochastic variable in an analysis"""
		return interface.GetValue(self._id, 'min_value')

	@min_value.setter
	def min_value(self, value : float):
		interface.SetValue(self._id, 'min_value', value)

	@property
	def max_value(self) -> float:
		"""The maximum value which will be assigned to the stochastic variable in an analysis"""
		return interface.GetValue(self._id, 'max_value')

	@max_value.setter
	def max_value(self, value : float):
		interface.SetValue(self._id, 'max_value', value)

	@property
	def start_value(self) -> float:
		"""The value of the starting point for this stochastic variable in an analysis"""
		return interface.GetValue(self._id, 'start_value')

	@start_value.setter
	def start_value(self, value : float):
		interface.SetValue(self._id, 'start_value', value)

	@property
	def intervals(self) -> int:
		"""The number of intervals for this stochastic variable, used by numerical integration"""
		return interface.GetIntValue(self._id, 'intervals')

	@intervals.setter
	def intervals(self, value : int):
		interface.SetIntValue(self._id, 'intervals', value)

	@property
	def variance_factor(self) -> float:
		"""The variance factor for this stochastic variable, used by importance sampling"""
		return interface.GetValue(self._id, 'variance_factor')

	@variance_factor.setter
	def variance_factor(self, value : float):
		interface.SetValue(self._id, 'variance_factor', value)

	@property
	def is_initialization_allowed(self) -> bool:
		"""Indicates whether the value in the starting point can be generated (true) or that the start_value is used (false)"""
		return interface.GetBoolValue(self._id, 'is_initialization_allowed')

	@is_initialization_allowed.setter
	def is_initialization_allowed(self, value : bool):
		interface.SetBoolValue(self._id, 'is_initialization_allowed', value)

	@property
	def is_variance_allowed(self) -> bool:
		"""Indicates whether the starting point is updated for this stochastic variable, used by adaptive importance sampling"""
		return interface.GetBoolValue(self._id, 'is_variance_allowed')

	@is_variance_allowed.setter
	def is_variance_allowed(self, value: bool):
		interface.SetBoolValue(self._id, 'is_variance_allowed', value)

class CompareType(Enum):
	"""Enumeration which defines a comparison type"""
	less_than = 'less_than'
	greater_than = 'greater_than'
	def __str__(self):
		return str(self.value)

class LimitStateFunction(FrozenObject):
	"""Defines how model output is transformed to a z-value, which is used by reliability analyses.

    A reliability algorithm only uses a z-value, where a value < 0 indicates failure and a value >= 0
    indicates no failure. A model produces several output values in the form of an array, where each
    array value is related to an output parameter of the model. This class allows the user to select
    an output parameter and compare it against a critical value.

    A limit state function should be added to a reliability project. If not added, the reliability
    algorithm uses the first value of the model output array and interprets it as the z-value."""

	def __init__(self, id = None):
		if id is None:
			self._id = interface.Create('limit_state_function')
		else:
			self._id = id
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['parameter',
		        'compare_type',
		        'critical_value']

	def __str__(self):
		return self.parameter + ' ' + str(self.compare_type) + ' ' + str(self.critical_value)

	@property
	def parameter(self) -> str:
		"""The output `probabilistic_library.project.ModelParameter` to be used"""
		return interface.GetStringValue(self._id, 'parameter')

	@parameter.setter
	def parameter(self, value : str | ModelParameter):
		interface.SetStringValue(self._id, 'parameter', str(value))

	@property
	def compare_type(self) -> CompareType:
		"""The way the output parameter is compared with a critical value (greater than, less than)"""
		return CompareType[interface.GetStringValue(self._id, 'compare_type')]

	@compare_type.setter
	def compare_type(self, value : CompareType):
		interface.SetStringValue(self._id, 'compare_type', str(value))

	@property
	def critical_value(self) -> float:
		"""The critical value with which the output parameter is compared
        This can be a numerical value or another model parameter (input or output)"""
		if interface.GetBoolValue(self._id, 'use_compare_parameter'):
			return interface.GetStringValue(self._id, 'compare_parameter')
		else:
			return interface.GetValue(self._id, 'critical_value')

	@critical_value.setter
	def critical_value(self, value : float | ModelParameter | str):
		if type(value) is float or type(value) is int:
			interface.SetBoolValue(self._id, 'use_compare_parameter', False)
			interface.SetValue(self._id, 'critical_value', value)
		else:
			interface.SetBoolValue(self._id, 'use_compare_parameter', True)
			interface.SetStringValue(self._id, 'compare_parameter', str(value))

class DesignPointIds(FrozenObject):
	"""Base class for additional design point information"""

	def __init__(self):
		pass # empty method as it is in a abstract class

class DesignPoint(FrozenObject):
	"""Result of a reliability analysis, containing the reliability index and contributions of
    stochastic variables.

    The main result of a reliability analysis is a reliability index, which can be transformed to a
    probability of failure. Also contributions of stochastic variables are given in the list of
    alphas.

    Convergence information, number of samples, iterations or directions are given too. Depending on
    settings, samples, messages and convergence information during the calculation are reported too.
    If intermediate design points were generated, for example by adaptive importance sampling, they
    are reported in the contributing design points list.

    The design point can further be used to combine it with other design points or upscale it to make
    a section design point applicable to a system"""

	def __init__(self, id = None, known_variables = None, known_design_points = None):
		if id is None:
			self._id = interface.Create('design_point')
		else:
			self._id = id

		self._alphas = None
		self._contributing_design_points = None
		self._messages = None
		self._realizations = None
		self._reliability_results = None
		self._ids = None
		self._known_variables = known_variables
		self._known_design_points = known_design_points
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['identifier',
		        'ids',
		        'reliability_index',
		        'probability_failure',
		        'alphas',
		        'contributing_design_points',
		        'convergence',
		        'is_converged',
		        'total_directions',
		        'total_iterations',
		        'total_model_runs',
		        'realizations',
		        'reliability_results',
		        'messages',
		        'print',
		        'plot_alphas',
		        'plot_realizations',
 		        'plot_convergence',
		        'get_plot_alphas',
		        'get_plot_realizations',
 		        'get_plot_convergence']

	def __str__(self):
		return self.identifier

	@property
	def identifier(self) -> str:
		"""Identifying text of the design point, generated by the reliability algorithm"""
		return interface.GetStringValue(self._id, 'identifier')

	@identifier.setter
	def identifier(self, value : str):
		interface.SetStringValue(self._id, 'identifier', value)

	@property
	def ids(self) -> DesignPointIds:
		"""Additional information"""
		return self._ids

	@ids.setter
	def ids(self, value : DesignPointIds):
		self._ids = value
		interface.SetIntValue(self._id, 'ids', value._id)

	@property
	def reliability_index(self) -> float:
		"""Reliability index, calculated by the reliability algorithm"""
		return interface.GetValue(self._id, 'reliability_index')

	# testing method
	def _set_reliability_index(self, reliability_index_value):
		interface.SetValue(self._id, 'reliability_index', reliability_index_value)

	# testing method
	def _add_alpha(self, variable, alpha_value):
		alpha = Alpha();
		alpha._set_alpha(variable, alpha_value, self.reliability_index);

		alphas = []
		alphas.extend(self.alphas)
		alphas.append(alpha)
		self._alphas = FrozenList(alphas)

		values = [a._id for a in self._alphas]
		interface.SetArrayIntValue(self._id, 'alphas', values)

	@property
	def probability_failure(self) -> float:
		"""Probability of failure, related to the reliability index"""
		return interface.GetValue(self._id, 'probability_failure')

	@property
	def convergence(self) -> float:
		"""Achieved convergence with the reliability algorithm"""
		return interface.GetValue(self._id, 'convergence')

	@property
	def is_converged(self) -> bool:
		"""Indicates whether convergence is reached"""
		return interface.GetBoolValue(self._id, 'is_converged')

	@property
	def total_directions(self) -> int:
		"""The total number of directions, used by the directional sampling algorithm"""
		return interface.GetIntValue(self._id, 'total_directions')

	@property
	def total_iterations(self) -> int:
		"""The total number of iterations, used by FORM"""
		return interface.GetIntValue(self._id, 'total_iterations')

	@property
	def total_model_runs(self) -> int:
		"""The total number of samples"""
		return interface.GetIntValue(self._id, 'total_model_runs')

	@property
	def alphas(self) -> list[Alpha]:
		"""List of contributions per stochastic variable to this design point

        The alphas indicate the contribution of the uncertainty of the stochastic variable to the design point.
        They also contain the value of the variable in the design point.

        For each variable in the reliability project variables an alpha variable is present in this list. In case
        of an array variable, for each array entry an alpha value is present."""

		if self._alphas is None:
			alphas = []
			alpha_ids = interface.GetArrayIdValue(self._id, 'alphas')
			for alpha_id in alpha_ids:
				alphas.append(Alpha(alpha_id, self._known_variables))
			self._alphas = FrozenList(alphas)
		return self._alphas

	@property
	def contributing_design_points(self) -> list[DesignPoint]:
		"""List of intermediate design points

        They can refer to loops in importance sampling or in FORM, when several relaxation loops
        are allowed. When a starting point was used and generated, this will also be part of the
        contributing design point."""

		if self._contributing_design_points is None:
			contributing_design_points = []
			design_point_ids = interface.GetArrayIdValue(self._id, 'contributing_design_points')
			for design_point_id in design_point_ids:
				if design_point_id > 0:
					added = False
					if not self._known_design_points is None:
						for design_point in self._known_design_points:
							if design_point._id == design_point_id:
								contributing_design_points.append(design_point)
								added = True

					if not added:
						contributing_design_points.append(DesignPoint(design_point_id, self._known_variables, self._known_design_points))
			self._contributing_design_points = FrozenList(contributing_design_points)

		return self._contributing_design_points

	@property
	def realizations(self) -> list[Evaluation]:
		"""List of samples calculated by the reliability algorithm. Depends on the setting `Settings.save_realizations` whether
        this list is provided"""
		if self._realizations is None:
			realizations = []
			realization_ids = interface.GetArrayIdValue(self._id, 'evaluations')
			for realization_id in realization_ids:
				realizations.append(Evaluation(realization_id))
			self._realizations = FrozenList(realizations)

		return self._realizations

	@property
	def reliability_results(self) -> list[ReliabilityResult]:
		"""List of convergence reports during the reliability analysis. Depends on the setting `Settings.save_convergence`
        whether this list is provided"""
		if self._reliability_results is None:
			reliability_results = []
			reliability_result_ids = interface.GetArrayIdValue(self._id, 'reliability_results')
			for reliability_result_id in reliability_result_ids:
				reliability_results.append(ReliabilityResult(reliability_result_id))
			self._reliability_results = FrozenList(reliability_results)

		return self._reliability_results

	@property
	def messages(self) -> list[Message]:
		"""List of messages generated by the reliability algorithm. Depends on the setting `Settings.save_messages` whether this
        list is provided"""
		if self._messages is None:
			messages = []
			message_ids = interface.GetArrayIdValue(self._id, 'messages')
			for message_id in message_ids:
				messages.append(Message(message_id))
			self._messages = FrozenList(messages)

		return self._messages

	def get_variables(self) -> list[Stochast]:
		"""Gets a list of all stochastic variables in this design point or contributing design points"""
		variables = []
		for alpha in self.alphas:
			variables.append(alpha.variable)
		for contributing_design_point in self.contributing_design_points:
			variables.extend(contributing_design_point.get_variables())
		return frozenset(variables)

	def print(self, decimals = 4):
		"""Prints the design point, including convergence, list of alphas and contributing design points

        Parameters
        ----------
        decimals : int, optional
            The number of decimals to print"""

		self._print(0, decimals)

	def _print(self, indent : int, decimals = 4):
		pre = PrintUtils.get_space_from_indent(indent)
		pre_indexed = pre + ' '
		if self.identifier == '':
			print(pre + 'Reliability:')
		else:
			print(pre + f'Reliability ({self.identifier})')
		print(pre_indexed + f'Reliability index = {self.reliability_index:.{decimals}g}')
		print(pre_indexed + f'Probability of failure = {self.probability_failure:.{decimals}g}')
		if not isnan(self.convergence):
			if self.is_converged:
				print(pre_indexed + f'Convergence = {self.convergence:.{decimals}g} (converged)')
			else:
				print(pre_indexed + f'Convergence = {self.convergence:.{decimals}g} (not converged)')
		print(pre_indexed + f'Model runs = {self.total_model_runs}')
		print(pre + 'Alpha values:')
		for alpha in self.alphas:
			print(pre_indexed + f'{alpha.identifier}: alpha = {alpha.alpha:.{decimals}g}, x = {alpha.x:.{decimals}g}')
		print('')
		if len(self.contributing_design_points) > 0:
			print(pre + 'Contributing design points:')
			for design_point in self.contributing_design_points:
				design_point._print(indent + 1, decimals)

	def plot_alphas(self):
		"""Shows a plot of the alphas in the form a pie-diagram"""
		self.get_plot_alphas().show()

	def get_plot_alphas(self) -> plt:
		"""Gets a plot object of the alphas in the form a pie-diagram"""

		import numpy as np

		alphas = [alpha.influence_factor for alpha in self.alphas if alpha.influence_factor > 0.0001]
		names = [f'{alpha.identifier} ({round(100*alpha.influence_factor)} %)' for alpha in self.alphas if alpha.influence_factor > 0.0001]

		plt.close()

		plt.figure()
		plt.pie(alphas, labels=names)
		plt.title("Squared alpha values", fontsize=14, fontweight='bold')

		return plt

	def plot_realizations(self, var_x : str | Stochast = None, var_y : str | Stochast = None):
		"""Shows a scatter-plot of realizations performed by the reliability analysis. The color
        indicates failure or non-failure. The x-y coordinates correspond with realization input
        values. Only available when `Settings.save_realizations` was set in the settings.

        Parameters
        ----------
        var_x : str | Stochast, optional
            The stochastic variable to use for the x-axis, if omitted the variable with the greatest
            influence factor is used

        var_y : str | Stochast, optional
            The stochastic variable to use for the y-axis, if omitted the variable with the one but
            greatest influence factor is used"""

		self.get_plot_realizations(var_x, var_y).show()

	def get_plot_realizations(self, var_x : str | Stochast = None, var_y : str | Stochast = None) -> plt:
		"""Gets a plot object of a scatter-plot of realizations performed by the reliability analysis. The color
        indicates failure or non-failure. The x-y coordinates correspond with realization input
        values. Only available when `Settings.save_realizations` was set in the settings.

        Parameters
        ----------
        var_x : str | Stochast, optional
            The stochastic variable to use for the x-axis, if omitted the variable with the greatest
            influence factor is used

        var_y : str | Stochast, optional
            The stochastic variable to use for the y-axis, if omitted the variable with the one but
            greatest influence factor is used"""

		if len(self.realizations) == 0:
			print ("No realizations were saved, run again with settings.save_realizations = True")

		if len(self.alphas) < 2:
			print ("Not enough variables to plot realizations")

		import numpy as np

		# 2 variables with the highest alpha
		alphas = [alpha.influence_factor for alpha in self.alphas]
		index_last_two = np.argsort(np.abs(alphas))[-2:]

		if var_x == None and var_y == None:
			index_x = int(index_last_two[0])
			index_y = int(index_last_two[1])
		if var_x != None and var_y == None:
			index_x = self.alphas.index(var_x)
			index_y = int(index_last_two[0])
		if var_x == None and var_y != None:
			index_x = int(index_last_two[0])
			index_y = self.alphas.index(var_y)
		if var_x != None and var_y != None:
			index_x = self.alphas.index(var_x)
			index_y = self.alphas.index(var_y)

		if index_x < 0 or index_y < 0:
			print ("Variables could not be found")

		x_values = [realization.input_values[index_x] for realization in self.realizations]
		y_values = [realization.input_values[index_y] for realization in self.realizations]
		z = [realization.z for realization in self.realizations]
		colors = ["r" if val < 0 else "g" for val in z]

		plt.close()

		# plot realizations
		plt.figure()
		plt.grid(True)
		plt.scatter(x_values, y_values, color=colors, alpha=0.5)
		plt.scatter(self.alphas[index_x].x,
                    self.alphas[index_y].x,
                    label='design point' if self.identifier == '' else self.identifier,
                    color="black")
		plt.xlabel(self.alphas[index_x].identifier)
		plt.ylabel(self.alphas[index_y].identifier)
		plt.legend()
		plt.title('Realizations: Red = Failure, Green = No Failure', fontsize=14, fontweight='bold')

		return plt

	def plot_convergence(self):
		"""Shows a plot of the convergence against the iteration or sample index.
        Only available when `Settings.save_convergence` was set"""
		self.get_plot_convergence().show()

	def get_plot_convergence(self) -> plt:
		"""Gets a plot object of the convergence against the iteration or sample index.
        Only available when `Settings.save_convergence` was set"""

		if len(self.reliability_results) == 0:
			print ("No convergence data were saved, run again with settings.save_convergence = True")

		import numpy as np

		index = [x.index for x in self.reliability_results]
		beta = [x.reliability_index for x in self.reliability_results]
		conv = [x.convergence for x in self.reliability_results]

		plt.close()

		ax1 = plt.subplot()
		color = "tab:blue"
		ax1.set_xlabel("index [-]")
		ax1.set_ylabel("reliability index [-]", color=color)
		ax1.plot(index, beta)
		ax1.tick_params(axis="y", labelcolor=color)
		ax2 = ax1.twinx()
		color = "tab:red"
		ax2.set_ylabel("convergence [-]", color=color)
		ax2.plot(index, conv, "r--", label="convergence")
		ax2.tick_params(axis="y", labelcolor=color)
		plt.title('Convergence', fontsize=14, fontweight='bold')

		return plt

class Alpha(FrozenObject):
	"""Contribution of a stochastic variable to a design point

    The contribution consists of two indicators, the influence of the uncertainty of the stochastic
    variable to the design point and the physical value of the variable in the design point."""

	def __init__(self, id = None, known_variables = None):
		if id is None:
			self._id = interface.Create('alpha')
		else:
			self._id = id

		self._variable = None
		self._known_variables = known_variables
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['variable',
				'identifier',
				'alpha',
				'alpha_correlated',
				'influence_factor',
				'index',
				'x',
				'u']

	@property
	def variable(self) -> Stochast:
		"""Stochastic variable to which this alpha value refers"""
		if self._variable is None:
			variable_id = interface.GetIdValue(self._id, 'variable')
			if variable_id > 0:
				if not self._known_variables is None:
					for variable in self._known_variables:
						if variable._id == variable_id:
							self._variable = variable

				if self._variable is None:
					self._variable = Stochast(variable_id);

		return self._variable

	def __str__(self):
		return self.identifier

	@property
	def identifier(self) -> str:
		"""Identifying string of the variable or, in case of an array variable, accompanied with its index number"""
		return interface.GetStringValue(self._id, 'identifier')

	# internal method
	def _set_variable(self, variable):
		self._variable = variable

	# testing method
	def _set_alpha(self, variable, alpha_value, beta):
		self._variable = variable
		interface.SetIntValue(self._id, 'variable', variable._id)
		interface.SetValue(self._id, 'alpha', alpha_value)

		u = - beta * alpha_value
		interface.SetValue(self._id, 'u',- beta * alpha_value)
		interface.SetValue(self._id, 'x', variable.get_x_from_u(u))

	@property
	def alpha(self) -> float:
		"""Alpha value, the contribution of the uncertainty of the variable"""
		return interface.GetValue(self._id, 'alpha')

	@property
	def alpha_correlated(self) -> float:
		"""Alpha value after transformation for correlation, used to determine the physical value 'x'"""
		return interface.GetValue(self._id, 'alpha_correlated')

	@property
	def influence_factor(self) -> float:
		"""Influence factor of the uncertainty of the variable
        This value equals alpha squared. All influence factors of a design point sum up to 1."""
		return interface.GetValue(self._id, 'influence_factor')

	@property
	def index(self) -> int:
		"""Index number of the variable, if it is defined as an array variable"""
		return interface.GetIntValue(self._id, 'index')

	@property
	def u(self) -> float:
		"""U-value of the variable in the design point"""
		return interface.GetValue(self._id, 'u')

	@property
	def x(self) -> float:
		"""Physical value of the variable in the design point"""
		return interface.GetValue(self._id, 'x')


class FragilityCurve(FrozenObject):
	"""Curve with a number of x-values, for which a reliability is defined

    A fragility curve is similar to a `probabilistic_library.statistic.Stochast` with distribution type
    cdf_curve. A fragility curve has an additional method `integrate` to calculate the reliability, when
    the distribution of the x-values is known.

    ```mermaid
    classDiagram
        class FragilityValue{
            +x: float
            +reliability_index: float
            +design_point : DesignPoint
        }
        class DesignPoint{
        }
        class FragilityCurve{
            +"derived properties" : float
            +integrand : Stochast
            +integrate() : DesignPoint
        }
        class FragilityCurveProject{
            +integrand Stochast
            +fragility_curve FragilityCurve
            +run()
        }
        FragilityValue "*" <-- FragilityCurve
        DesignPoint <-- FragilityValue
        DesignPoint <-- FragilityCurveProject
        FragilityCurve <-- FragilityCurveProject
    ```
    """

	def __init__(self, id = None):
		if id is None:
			self._id = interface.Create('fragility_curve')
		else:
			self._id = id

		self._fragility_values = None
		self._synchronizing = False
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['name',
				'mean',
				'deviation',
				'variation',
				'fragility_values',
				'copy_from',
				'get_quantile',
				'get_x_from_u',
				'get_u_from_x',
				'get_pdf',
				'get_cdf',
				'get_special_values',
				'integrate']

	@property
	def name(self) -> str:
		"""Identifying name of the fragility curve"""
		return interface.GetStringValue(self._id, 'name')

	@name.setter
	def name(self, value : str):
		interface.SetStringValue(self._id, 'name', value)

	def __str__(self):
		return self.name

	@property
	def mean(self) -> float:
		"""Gets the mean of the fragility curve"""
		return interface.GetValue(self._id, 'mean')

	@property
	def deviation(self) -> float:
		"""Gets the standard deviation of the fragility curve"""
		return interface.GetValue(self._id, 'deviation')

	@property
	def variation(self) -> float:
		"""Gets the variation coefficient of the fragility curve"""
		return interface.GetValue(self._id, 'variation')

	@property
	def fragility_values(self) -> list[FragilityValue]:
		"""List of fragility values. Each fragility value defines a reliability index for an x-value.

        Preferably the design point from which the reliability index is taken should be specified too,
        which improves the `integrate` calculation"""

		if self._fragility_values is None:
			self._synchronizing = True
			self._fragility_values = CallbackList(self._fragility_values_changed)
			fragility_ids = interface.GetArrayIdValue(self._id, 'fragility_values')
			for fragility_id in fragility_ids:
				self._fragility_values.append(FragilityValue(fragility_id))
			self._synchronizing = False

		return self._fragility_values

	def _fragility_values_changed(self):
		if not self._synchronizing:
			interface.SetArrayIntValue(self._id, 'fragility_values', [fragility_value._id for fragility_value in self._fragility_values])

	def get_quantile(self, quantile : float) -> float:
		"""Gets the value belonging to a given quantile

        Parameters
        ----------
        quantile : float.
            Quantile for which the x-value is requested, must be between 0 and 1 (exclusive)"""

		return interface.GetArgValue(self._id, 'quantile', quantile)

	def get_x_from_u(self, u : float) -> float:
		"""Gets the x-value at a given u-value of the fragility curve

        Parameters
        ----------
        u : float.
            U-value for which the x-value is requested"""

		return interface.GetArgValue(self._id, 'x_from_u', u)

	def get_u_from_x(self, x : float) -> float:
		"""Gets the u-value at a given x-value of the fragility curve

        Parameters
        ----------
        x : float
            X-value for which the u-value is requested"""

		return interface.GetArgValue(self._id, 'u_from_x', x)

	def get_pdf(self, x : float) -> float:
		"""Gets the PDF value of the fragility curve

        Parameters
        ----------
        x : float
            X-value for which the PDF is requested"""

		return interface.GetArgValue(self._id, 'pdf', x)

	def get_cdf(self, x : float) -> float:
		"""Gets the CDF value of the fragility curve

        Parameters
        ----------
        x : float
            X-value for which the CDF is requested"""

		return interface.GetArgValue(self._id, 'cdf', x)

	def get_special_values(self) -> list[float]:
		"""Gets a list of special x-values, which is useful for plotting"""
		return interface.GetArrayValue(self._id, 'special_values')

	def copy_from(self, source):
		"""Copies the properties from a source fragility curve.

        Parameters
        ----------
        source : FragilityCurve
            The fragility curve to copy the properties from"""

		if source is FragilityCurve:
			interface.SetIntValue(self._id, 'copy_from', source._id)
			self._fragility_values = None

	def integrate(self, integrand : Stochast) -> DesignPoint:
		"""Calculates the reliability index of the fragility curve if the distribution of the x-values in the
        `fragility_values` is given. Results in a design point.

        The calculation is performed by integration over the possible x-values. If the design points in the
        fragility values are given, the alpha values in the resulting design point are based on the design
        points in the `fragility_values`.

        Parameters
        ----------
        integrand : Stochast.
            Stochastic variable describing the distribution of the x-values in the 'fragility_values'"""

		project = FragilityCurveProject()
		project.integrand = integrand
		project.fragility_curve = self
		project.run()
		return project.design_point

class FragilityCurveProject(FrozenObject):
	"""Project to calculate the reliability of a fragility curve when the distribution of the x-values of
    the fragility_valus of the fragility curve is known.

    This class is used internally by the `FragilityCurve.integrate` method."""

	def __init__(self):
		self._id = interface.Create('fragility_curve_project')
		self._integrand = None
		self._fragility_curve = None
		self._design_point = None
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['integrand',
				'fragility_curve',
				'design_point']

	@property
	def integrand(self) -> Stochast:
		"""Stochastic variable describing the distribution of the x-values in the `FragilityCurve.fragility_values`"""
		return self._integrand

	@integrand.setter
	def integrand(self, value : Stochast):
		self._integrand = value
		interface.SetIntValue(self._id, 'integrand', value._id)

	@property
	def fragility_curve(self) -> FragilityCurve:
		"""fragility curve to be integrated"""
		return self._fragility_curve

	@fragility_curve.setter
	def fragility_curve(self, value : FragilityCurve):
		self._fragility_curve = value
		interface.SetIntValue(self._id, 'fragility_curve', value._id)

	@property
	def design_point(self) -> DesignPoint:
		"""Resulting design point"""
		return self._design_point

	def run(self):
		"""Performs the integration"""
		self._design_point = None
		interface.Execute(self._id, 'run')
		design_point_id = interface.GetIdValue(self._id, 'design_point')
		if design_point_id > 0:
			known_variables = self._get_variables()
			self._design_point = DesignPoint(design_point_id, known_variables)

	def _get_variables(self):
		variables = []
		variables.append(self.integrand)
		variables.append(self.fragility_curve)
		for fragility_value in self.fragility_curve.fragility_values:
			if isinstance(fragility_value.design_point, DesignPoint):
				for alpha in fragility_value.design_point.alphas:
					if alpha.variable is not None and alpha.variable not in variables:
						variables.append(alpha.variable)
						for array_variable in alpha.variable.array_variables:
							if array_variable not in variables:
								variables.append(array_variable)
		return variables


class CombineSettings(FrozenObject):
	"""Settings for combining design points
    These settings are used by the `probabilistic_library.project.CombineProject`"""

	def __init__(self):
		self._id = interface.Create('combine_settings')
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['combiner_method',
				'combine_type']

	@property
	def combiner_method(self) -> CombinerMethod:
		"""Algorithm which performs the combination of design points"""
		return CombinerMethod[interface.GetStringValue(self._id, 'combiner_method')]

	@combiner_method.setter
	def combiner_method(self, value : CombinerMethod):
		interface.SetStringValue(self._id, 'combiner_method', str(value))

	@property
	def combine_type(self) -> CombineType:
		"""Type of combining design points (series, parallel)"""
		return CombineType[interface.GetStringValue(self._id, 'combine_type')]

	@combine_type.setter
	def combine_type(self, value : CombineType):
		interface.SetStringValue(self._id, 'combine_type', str(value))

class ExcludingCombineSettings(FrozenObject):
	"""Settings for combining design points exclusively
    These settings are used by the `probabilistic_library.project.ExcludingCombineProject`"""

	def __init__(self):
		self._id = interface.Create('excluding_combine_settings')
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['combiner_method']

	@property
	def combiner_method(self) -> ExcludingCombinerMethod:
		"""Algorithm which performs the exclusive combination of design points"""
		return ExcludingCombinerMethod[interface.GetStringValue(self._id, 'combiner_method')]

	@combiner_method.setter
	def combiner_method(self, value : ExcludingCombinerMethod):
		interface.SetStringValue(self._id, 'combiner_method', str(value))


class ReliabilityResult(FrozenObject):
	"""Contains the intermediate reliability index and convergence during a reliability analysis"""

	def __init__(self, id = None):
		if id == None:
			self._id = interface.Create('reliability_result')
		else:
			self._id = id
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['index',
				'reliability_index',
				'convergence']

	@property
	def index(self) -> int:
		"""Index of a realization or iteration in the reliability analysis to which the reliability index and convergence belong"""
		return interface.GetIntValue(self._id, 'index')

	@property
	def reliability_index(self) -> float:
		"""Intermediate reliability index in the reliability analysis"""
		return interface.GetValue(self._id, 'reliability_index')

	@property
	def convergence(self) -> float:
		"""Intermediate convergence index in the reliability analysis"""
		return interface.GetValue(self._id, 'convergence')



# Copyright (C) Stichting Deltares. All rights reserved.
#
# This file is part of the Probabilistic Library.
#
# The Probabilistic Library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# All names, logos, and references to "Deltares" are registered trademarks of
# Stichting Deltares and remain full property of Stichting Deltares at all times.
# All rights reserved.
#
from __future__ import annotations
from math import isnan
from enum import Enum

from .utils import FrozenObject, FrozenList, PrintUtils
from . import interface

if not interface.IsLibraryLoaded():
	interface.LoadDefaultLibrary()

class MessageType(Enum):
	"""Enumeration which defines the severity of a message."""
	debug = 'debug'
	info = 'info'
	warning = 'warning'
	error = 'error'
	def __str__(self):
		return str(self.value)

class Message(FrozenObject):
	"""Defines a message, which will be presented to the user.
    A message may be the result of a validation request or an informative message of an algorithm"""

	def __init__(self, id = None):
		if id == None:
			self._id = interface.Create('message')
		else:
			self._id = id
		super()._freeze()

	@classmethod
	def from_message(cls, message_type, message_text):
		message = cls()
		interface.SetStringValue(message._id, 'type', str(message_type))
		interface.SetStringValue(message._id, 'text', message_text)
		return message

	def __del__(self):
		interface.Destroy(self._id)
		
	def __str__(self):
		if self.subject == "":
			return str(self.type) + ': ' + self.text
		else:
			return str(self.type) + ': ' + self.subject + ' => ' + self.text
		
	def __dir__(self):
		return ['type',
		        'subject',
		        'text',
		        'print']

	@property
	def type(self) -> MessageType:
		"""Gets the message type, which indicates the severity of the message"""
		return MessageType[interface.GetStringValue(self._id, 'type')]
		
	@property
	def subject(self) -> str:
		"""Gets the subject, the object to which the message applies"""
		return interface.GetStringValue(self._id, 'subject')

	@property
	def text(self) -> str:
		"""Gets the message text"""
		return interface.GetStringValue(self._id, 'text')

	def print(self):
		"""Prints the message"""
		text = str(self)
		if len(text) > 0:
			text = text[0].capitalize() + text[1:]
		print(text)

class ValidationReport(FrozenObject):
	"""Result of a validation analysis, consists of a number of validation messages"""

	def __init__(self, id = None):
		if id == None:
			self._id = interface.Create('validation_report')
		else:
			self._id = id
		self._messages = None
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['is_valid',
		        'messages',
		        'print']

	def is_valid(self) -> bool:
		"""Indicates whether this validation report should be interpreted as valid or non-valid"""
		return interface.GetBoolValue(self._id, 'is_valid')

	@property   
	def messages(self) -> FrozenList[Message]:
		"""List of all validation messages"""
		if self._messages is None:
			message_ids = interface.GetArrayIdValue(self._id, 'messages')
			messages = []
			for message_id in message_ids:
				message = Message(message_id)
				messages.append(message)
			self._messages = FrozenList(messages)
		return self._messages

	def print(self):
		"""Prints all validation messages or an indication in case of no validation messages"""
		if len(self.messages) == 0:
			print('ok')
		else:
			for message in self.messages:
				message.print()

		
class Evaluation(FrozenObject):
	"""Registers a sample and execution results of a model"""
		
	def __init__(self, id = None):
		if id == None:
			self._id = interface.Create('evaluation')
		else:
			self._id = id
		self._input_values = None	
		self._output_values = None	
		super()._freeze()

	def __del__(self):
		interface.Destroy(self._id)

	def __dir__(self):
		return ['iteration',
				'quantile',
				'z',
				'beta',
				'weight',
				'input_values',
				'output_values',
				'print']
	
	@property   
	def iteration(self) -> int:
		"""Iteration index in a reliability, uncertainty or sensitivity algorithm"""
		return interface.GetIntValue(self._id, 'iteration')
		
	@property   
	def quantile(self) -> float:
		"""Quantile to which the sample belongs in an uncertainty algorithm"""
		return interface.GetValue(self._id, 'quantile')
		
	@property   
	def z(self) -> float:
		"""Z-value, indicating failure or non-failure, used by a reliability algorithm"""
		return interface.GetValue(self._id, 'z')
		
	@property   
	def beta(self) -> float:
		"""Distance of the sample to the origin in u-space"""
		return interface.GetValue(self._id, 'beta')
		
	@property   
	def weight(self) -> float:
		"""Weight of the sample in the reliability or uncertainty algorithm"""
		return interface.GetValue(self._id, 'weight')

	@property   
	def input_values(self) -> FrozenList[float]:
		"""List of input values for the model"""
		if self._input_values is None:
			input_values = interface.GetArrayValue(self._id, 'input_values')
			self._input_values = FrozenList(input_values)
		return self._input_values
		
	@property   
	def output_values(self) -> FrozenList[float]:
		"""List of output values produced by the model"""
		if self._output_values is None:
			output_values = interface.GetArrayValue(self._id, 'output_values')
			self._output_values = FrozenList(output_values)
		return self._output_values

	def print(self, decimals = 4):
		"""Prints the evaluation

        Parameters
        ----------
        decimals : int, optional.
            The number of decimals to print"""

		self._print(0, decimals)

	def _print(self, indent, decimals = 4):
		pre = PrintUtils.get_space_from_indent(indent)
		input_values = ', '.join([f'{v:.{decimals}g}' for v in self.input_values])
		output_values = ', '.join([f'{v:.{decimals}g}' for v in self.output_values])
		if not isnan(self.quantile):
			pre = pre + f'quantile {self.quantile:.{decimals}g}: '
		if isnan(self.z) and len(self.output_values) == 0:
			print(pre + f'[{input_values}]')
		elif isnan(self.z) and len(self.output_values) > 0:
			print(pre + f'[{input_values}] -> [{output_values}]')
		elif not isnan(self.z) and len(self.output_values) == 0:
			print(pre + f'[{input_values}] -> {self.z:.{decimals}g}')
		elif not isnan(self.z) and len(self.output_values) > 0:
			print(pre + f'[{input_values}] -> [{output_values}] -> {self.z:.{decimals}g}')
		
